from unittest import TestCase

from dark.dna import (
    AMBIGUOUS,
    BASES_TO_AMBIGUOUS,
    Bases,
    FloatBaseCounts,
    compareDNAReads,
    findKozakConsensus,
    leastAmbiguous,
    leastAmbiguousFromCounts,
    matchToString,
    sequenceToRegex,
)
from dark.reads import DNAKozakRead, DNARead, Read

# The following are the letters that used to be on
# from Bio.Alphabet.IUPAC import IUPACAmbiguousDNA
# IUPACAmbiguousDNA.letters
# But Bio.Alphabet is now deprecated and will be removed.
AMBIGUOUS_DNA_LETTERS = "GATCRYWSMKHBVDN"

AMBIGUOUS_PAIRS = (
    ("AC", "M"),
    ("AG", "R"),
    ("AT", "W"),
    ("GC", "S"),
    ("GT", "K"),
    ("CT", "Y"),
)

AMBIGUOUS_TRIPLES = (("ACG", "V"), ("ACT", "H"), ("AGT", "D"), ("CGT", "B"))


class TestAmbiguousLetters(TestCase):
    """
    Test the ambiguous DNA mapping.
    """

    def testExpectedAmbiguousLetters(self):
        """
        The ambiguous DNA letters must match those given by IUPAC.
        """
        self.assertEqual(sorted(AMBIGUOUS_DNA_LETTERS), sorted(AMBIGUOUS))

    def testExpectedLengthOne(self):
        """
        The unambiguous DNA letters must be in sets by themselves.
        """
        for base in "ACGT":
            self.assertEqual({base}, AMBIGUOUS[base])

    def testExpectedLengthsGreaterThanOne(self):
        """
        The ambiguous DNA letters must be in sets of size greater than one
        and less than 5.
        """
        for base in set(AMBIGUOUS_DNA_LETTERS) - set("ACGT"):
            self.assertTrue(5 > len(AMBIGUOUS[base]) > 1)

    def testAmbiguousLettersAreAllACGT(self):
        """
        The ambiguous DNA letter sets must all be drawn from A, C, G, T.
        """
        for bases in AMBIGUOUS.values():
            self.assertEqual(set(), bases - set("ACGT"))

    def testAmbiguousLettersAreAllSets(self):
        """
        The ambiguous DNA letter sets must be of the set type (this test
        is in lieu of testing that a tuple or list has no repeats).
        """
        for bases in AMBIGUOUS.values():
            self.assertTrue(isinstance(bases, set))


class TestReversedAmbiguousLetters(TestCase):
    """
    Test the reversed ambiguous DNA mapping.
    """

    def testExpectedSingleBase(self):
        """
        The unambiguous bases must map to themselves.
        """
        for base in "ACGT":
            self.assertEqual(base, BASES_TO_AMBIGUOUS[base])

    def testNIsAllBases(self):
        """
        N must map to all bases.
        """
        self.assertEqual("N", BASES_TO_AMBIGUOUS["ACGT"])


class TestCompareDNAReads(TestCase):
    """
    Test the compareDNAReads function.
    """

    def testEmptySequences(self):
        """
        Two empty sequences must compare as expected.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 0,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(Read("id1", ""), Read("id2", "")),
        )

    def testExactMatch(self):
        """
        Two sequences that match exactly must compare as expected.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(Read("id1", "ACGTT"), Read("id2", "ACGTT")),
        )

    def testNoOffsets(self):
        """
        If an empty set of wanted offsets is passed, the result must be empty.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 0,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(Read("id1", "ATT-T"), Read("id2", "A-GTC"), offsets=set()),
        )

    def testOffsets(self):
        """
        If a set of wanted offsets is passed, the result must be restricted to
        just those offsets.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 1,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 1,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [(4, "T", "C")],
                    "differenceOffsets": {
                        "TC": [4],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ATT-T"), Read("id2", "A-GTC"), offsets=set([0, 4])
            ),
        )

    def testMatchWithAmbiguityButStrict(self):
        """
        Two sequences that match exactly, apart from one ambiguity in the first
        sequence, must compare as expected when we specify matchAmbiguous=False
        to disallow ambiguous matching.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 1,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [(5, "S", "C")],
                    "differenceOffsets": {
                        "SC": [5],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTTS"),
                Read("id2", "ACGTTC"),
                matchAmbiguous=False,
            ),
        )

    def testMatchWithAmbiguityInFirst(self):
        """
        Two sequences that match exactly, apart from one ambiguity in the first
        sequence, must compare as expected.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 1,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [(5, "S", "C")],
                    "nonGapMismatches": [],
                    # There are no differences since the ambiguous S code matches
                    # the C.
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTTS"),
                Read("id2", "ACGTTC"),
            ),
        )

    def testMatchWithAmbiguityInSecond(self):
        """
        Two sequences that match exactly, apart from one ambiguity in the
        second sequence, must compare as expected.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 1,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [(5, "C", "S")],
                    "nonGapMismatches": [],
                    # There are no differences since the ambiguous S code matches
                    # the C.
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTTC"),
                Read("id2", "ACGTTS"),
            ),
        )

    def testNonMatchingAmbiguityInFirst(self):
        """
        Two sequences that match exactly, apart from one (incompatible)
        ambiguity in the second sequence, must compare as expected.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 1,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [(5, "W", "C")],
                    "differenceOffsets": {
                        "WC": [5],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTTW"),
                Read("id2", "ACGTTC"),
            ),
        )

    def testMatchWithAmbiguityInBoth(self):
        """
        Two sequences that match exactly, apart from one (compatible)
        ambiguity at the same location in the sequence, must compare as
        expected.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 1,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [(5, "K", "S")],
                    "nonGapMismatches": [],
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTTK"),
                Read("id2", "ACGTTS"),
            ),
        )

    def testMatchWithAmbiguityInBothButStrict(self):
        """
        Two sequences that match exactly, apart from one (compatible)
        ambiguity at the same location in the sequence, must compare as
        expected. Strict.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 1,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [(5, "K", "S")],
                    "differenceOffsets": {
                        "KS": [5],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTTK"),
                Read("id2", "ACGTTS"),
                matchAmbiguous=False,
            ),
        )

    def testMatchWithIncompatibleAmbiguityInBoth(self):
        """
        Two sequences that match exactly, apart from one (incompatible)
        ambiguity at the same location in the sequence, must compare as
        expected.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 1,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [(5, "W", "S")],
                    "differenceOffsets": {
                        "WS": [5],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(Read("id1", "ACGTTW"), Read("id2", "ACGTTS")),
        )

    def testMatchWithIncompatibleAmbiguityInBothButStrict(self):
        """
        Two sequences that match exactly, apart from one (incompatible)
        ambiguity at the same location in the sequence, must compare as
        expected. Strict.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 1,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [(5, "W", "S")],
                    "differenceOffsets": {
                        "WS": [5],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTTW"),
                Read("id2", "ACGTTS"),
                matchAmbiguous=False,
            ),
        )

    def testMatchWithIdenticalAmbiguity(self):
        """
        Two sequences that match exactly, including one (identical)
        ambiguity at the same location in the sequence, must compare as
        expected.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 1,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [(5, "N", "N")],
                    "nonGapMismatches": [],
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(Read("id1", "ACGTTN"), Read("id2", "ACGTTN")),
        )

    def testMatchWithIdenticalAmbiguityButStrict(self):
        """
        Two sequences that match exactly, including one (identical)
        ambiguity at the same location in the sequence, must compare as
        expected. Strict.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 1,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [(5, "N", "N")],
                    "differenceOffsets": {
                        # The two N's are not considered to match, because we
                        # pass matchAmbiguous=False to compareDNAReads.
                        "NN": [5],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [5],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTTN"), Read("id2", "ACGTTN"), matchAmbiguous=False
            ),
        )

    def testGapInFirst(self):
        """
        A gap in the first sequence must be dealt with correctly.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 4,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 1,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {
                        "-G": [2],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [2],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "AC-TT"),
                Read("id2", "ACGTT"),
            ),
        )

    def testGapInSecond(self):
        """
        A gap in the second sequence must be dealt with correctly.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 3,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 2,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {
                        "C-": [1],
                        "G-": [2],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [1, 2],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTT"),
                Read("id2", "A--TT"),
            ),
        )

    def testNonDefaultGapChars(self):
        """
        We must be able to specify the gap characters.
        """
        for gap in "+$":
            self.assertEqual(
                {
                    "match": {
                        "identicalMatchCount": 3,
                        "ambiguousMatchCount": 0,
                        "gapMismatchCount": 2,
                        "gapGapMismatchCount": 0,
                        "nonGapMismatchCount": 0,
                        "noCoverageCount": 0,
                        "noCoverageNoCoverageCount": 0,
                        "ambiguousMatches": [],
                        "nonGapMismatches": [],
                        "differenceOffsets": {
                            f"A{gap}": [0],
                            f"{gap}G": [2],
                        },
                    },
                    "read1": {
                        "ambiguousOffsets": [],
                        "extraCount": 0,
                        "gapOffsets": [2],
                        "noCoverageOffsets": [],
                    },
                    "read2": {
                        "ambiguousOffsets": [],
                        "extraCount": 0,
                        "gapOffsets": [0],
                        "noCoverageOffsets": [],
                    },
                },
                compareDNAReads(
                    Read("id1", "AC%sTT" % gap),
                    Read("id2", "%sCGTT" % gap),
                    gapChars="+$",
                ),
            )

    def testGapGap(self):
        """
        Coinciding gaps in the sequences must be dealt with correctly.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 2,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 2,
                    "gapGapMismatchCount": 1,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {
                        "C-": [1],
                        "-T": [3],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [2, 3],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [1, 2],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "AC--T"),
                Read("id2", "A--TT"),
            ),
        )

    def testGapAmbiguous(self):
        """
        Testing that the ambiguousOffset shows ambiguous characters paired
        with gaps as expected
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 2,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 2,
                    "gapGapMismatchCount": 1,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {
                        "N-": [1],
                        "-N": [3],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [1],
                    "extraCount": 0,
                    "gapOffsets": [2, 3],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [3],
                    "extraCount": 0,
                    "gapOffsets": [1, 2],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "AN--T"),
                Read("id2", "A--NT"),
            ),
        )

    def testNoCoverageInFirst(self):
        """
        A no coverage character in the first sequence must be dealt with
        correctly.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 4,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 1,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [2],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "AC?TT"), Read("id2", "ACGTT"), noCoverageChars="?"
            ),
        )

    def testNoCoverageInSecond(self):
        """
        A no coverage in the second sequence must be dealt with correctly.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 3,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 2,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [1, 2],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTT"), Read("id2", "A??TT"), noCoverageChars="?"
            ),
        )

    def testNoCoverageNoCoverage(self):
        """
        Coinciding no coverages in the sequences must be dealt with correctly.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 2,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 2,
                    "noCoverageNoCoverageCount": 1,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [2, 3],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [1, 2],
                },
            },
            compareDNAReads(
                Read("id1", "AC??T"),
                Read("id2", "A??TT"),
                noCoverageChars="?",
            ),
        )

    def testNoCoverageAndGaps(self):
        """
        When no coverage and gap characters are both found in the sequences,
        the result must be correct.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 2,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 3,
                    "gapGapMismatchCount": 2,
                    "nonGapMismatchCount": 1,
                    "noCoverageCount": 2,
                    "noCoverageNoCoverageCount": 1,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [(10, "T", "A")],
                    "differenceOffsets": {
                        "-T": [7],
                        "C-": [8],
                        "G-": [9],
                        "TA": [10],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [5, 6, 7],
                    "noCoverageOffsets": [2, 3],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [5, 6, 8, 9],
                    "noCoverageOffsets": [1, 2],
                },
            },
            compareDNAReads(
                Read("id1", "AC??T---CGT"),
                Read("id2", "A??TT--T--A"),
                noCoverageChars="?",
            ),
        )

    def testExtraInFirst(self):
        """
        If the first sequence has extra bases, they must be indicated in the
        extraCount.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 2,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTTCC"),
                Read("id2", "ACGTT"),
            ),
        )

    def testExtraInSecond(self):
        """
        If the second sequence has extra bases, they must be indicated in the
        extraCount.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 2,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTT"),
                Read("id2", "ACGTTCC"),
            ),
        )

    def testExtraAmbiguous(self):
        """
        If the first sequence has extra bases which are ambiguous,they must
        be indicated in the extraCount and in the ambiguousOffset.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 5,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 0,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [],
                    "differenceOffsets": {},
                },
                "read1": {
                    "ambiguousOffsets": [6],
                    "extraCount": 2,
                    "gapOffsets": [5],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTT-N"),
                Read("id2", "ACGTT"),
            ),
        )

    def testMismatch(self):
        """
        If the sequences have mismatched (non-ambiguous) bases, their count
        must be given correctly in the nonGapMismatchCount.
        """
        self.assertEqual(
            {
                "match": {
                    "identicalMatchCount": 3,
                    "ambiguousMatchCount": 0,
                    "gapMismatchCount": 0,
                    "gapGapMismatchCount": 0,
                    "nonGapMismatchCount": 2,
                    "noCoverageCount": 0,
                    "noCoverageNoCoverageCount": 0,
                    "ambiguousMatches": [],
                    "nonGapMismatches": [(3, "T", "C"), (4, "T", "C")],
                    "differenceOffsets": {
                        "TC": [3, 4],
                    },
                },
                "read1": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
                "read2": {
                    "ambiguousOffsets": [],
                    "extraCount": 0,
                    "gapOffsets": [],
                    "noCoverageOffsets": [],
                },
            },
            compareDNAReads(
                Read("id1", "ACGTT"),
                Read("id2", "ACGCC"),
            ),
        )


class TestMatchToString(TestCase):
    """
    Test the matchToString function.
    """

    def testMatchWithAmbiguityButStrict(self):
        """
        Two sequences that match exactly, apart from one ambiguity in the first
        sequence, must compare as expected when we specify matchAmbiguous=False
        to disallow ambiguous matching.
        """
        read1 = Read("id1", "ACGTTS")
        read2 = Read("id2", "ACGTTC")
        match = compareDNAReads(read1, read2, matchAmbiguous=False)

        self.assertEqual(
            """\
Exact matches: 5/6 (83.33%)
Ambiguous matches: 0
Mismatches: 1/6 (16.67%)
  Not involving gaps (i.e., conflicts or ambiguities): 1/6 (16.67%)
  Involving a gap in one sequence: 0
  Involving a gap in both sequences: 0
  Involving no coverage in one sequence: 0
  Involving no coverage in both sequences: 0
  Id: id1
    Length: 6
    Gaps: 0
    No coverage: 0
    Ambiguous: 1/6 (16.67%)
  Id: id2
    Length: 6
    Gaps: 0
    No coverage: 0
    Ambiguous: 0""",
            matchToString(match, read1, read2, matchAmbiguous=False),
        )

    def testMatchWithAmbiguityAndNotStrict(self):
        """
        Two sequences that match exactly, apart from one ambiguity in the first
        sequence, must compare as expected when we specify matchAmbiguous=True
        to allow ambiguous matching.
        """
        read1 = Read("id1", "ACGTTS")
        read2 = Read("id2", "ACGTTC")
        match = compareDNAReads(read1, read2, matchAmbiguous=True)

        self.assertEqual(
            """\
Exact matches: 5/6 (83.33%)
Ambiguous matches: 1/6 (16.67%)
Exact or ambiguous matches: 6/6 (100.00%)
Mismatches: 0
  Not involving gaps (i.e., conflicts): 0
  Involving a gap in one sequence: 0
  Involving a gap in both sequences: 0
  Involving no coverage in one sequence: 0
  Involving no coverage in both sequences: 0
  Id: id1
    Length: 6
    Gaps: 0
    No coverage: 0
    Ambiguous: 1/6 (16.67%)
  Id: id2
    Length: 6
    Gaps: 0
    No coverage: 0
    Ambiguous: 0""",
            matchToString(match, read1, read2, matchAmbiguous=True),
        )

    def testGapLocations(self):
        """
        Gap locations must be returned correctly.
        """
        read1 = Read("id1", "TTTTTAAAAAAGCGCG")
        read2 = Read("id2", "TTTTT------GCGCG")
        match = compareDNAReads(read1, read2)
        self.assertEqual(
            """\
Exact matches: 10/16 (62.50%)
Ambiguous matches: 0
Mismatches: 6/16 (37.50%)
  Not involving gaps (i.e., conflicts): 0
  Involving a gap in one sequence: 6/16 (37.50%)
  Involving a gap in both sequences: 0
  Involving no coverage in one sequence: 0
  Involving no coverage in both sequences: 0
  Id: id1
    Length: 16
    Gaps: 0
    No coverage: 0
    Ambiguous: 0
  Id: id2
    Length: 16
    Gaps: 6/16 (37.50%)
    Gap locations (1-based): 6, 7, 8, 9, 10, 11
    No coverage: 0
    Ambiguous: 0""",
            matchToString(match, read1, read2),
        )

    def testAmbiguousMatchLocations(self):
        """
        Ambiguous match locations must be returned correctly.
        """
        read1 = Read("id1", "TWTTTAAAAAAGCGCG")
        read2 = Read("id2", "TTTTT------GCGCK")
        match = compareDNAReads(read1, read2)
        self.assertEqual(
            """\
Exact matches: 8/16 (50.00%)
Ambiguous matches: 2/16 (12.50%)
Exact or ambiguous matches: 10/16 (62.50%)
Mismatches: 6/16 (37.50%)
  Not involving gaps (i.e., conflicts): 0
  Involving a gap in one sequence: 6/16 (37.50%)
  Involving a gap in both sequences: 0
  Involving no coverage in one sequence: 0
  Involving no coverage in both sequences: 0
  Id: id1
    Length: 16
    Gaps: 0
    No coverage: 0
    Ambiguous: 1/16 (6.25%)
  Id: id2
    Length: 16
    Gaps: 6/16 (37.50%)
    Gap locations (1-based): 6, 7, 8, 9, 10, 11
    No coverage: 0
    Ambiguous: 1/16 (6.25%)
Ambiguous matches:
    2 W T
    16 G K""",
            matchToString(match, read1, read2, includeAmbiguousMatches=True),
        )

    def testNonGapMismatchLocations(self):
        """
        Non-gap mismatch locations must be returned correctly.
        """
        read1 = Read("id1", "TATTTAAAAAAGCGCG")
        read2 = Read("id2", "TTTTT------GCGCC")
        match = compareDNAReads(read1, read2)
        self.assertEqual(
            """\
Exact matches: 8/16 (50.00%)
Ambiguous matches: 0
Mismatches: 8/16 (50.00%)
  Not involving gaps (i.e., conflicts): 2/16 (12.50%)
  Involving a gap in one sequence: 6/16 (37.50%)
  Involving a gap in both sequences: 0
  Involving no coverage in one sequence: 0
  Involving no coverage in both sequences: 0
  Id: id1
    Length: 16
    Gaps: 0
    No coverage: 0
    Ambiguous: 0
  Id: id2
    Length: 16
    Gaps: 6/16 (37.50%)
    Gap locations (1-based): 6, 7, 8, 9, 10, 11
    No coverage: 0
    Ambiguous: 0
Non-gap mismatches:
    2 A T
    16 G C""",
            matchToString(match, read1, read2, includeNonGapMismatches=True),
        )

    def testNoCoverageLocations(self):
        """
        No coverage locations must be returned correctly.
        """
        read1 = Read("id1", "TTTTTAAAAAAGCGCG")
        read2 = Read("id2", "TTTTT??????GCGCG")
        match = compareDNAReads(read1, read2, noCoverageChars="?")
        self.assertEqual(
            """\
Exact matches: 10/16 (62.50%)
Ambiguous matches: 0
Exact matches (ignoring no coverage sites): 10/10 (100.00%)
Ambiguous matches (ignoring no coverage sites): 0
Mismatches: 0
  Not involving gaps (i.e., conflicts): 0
  Involving a gap in one sequence: 0
  Involving a gap in both sequences: 0
  Involving no coverage in one sequence: 6/16 (37.50%)
  Involving no coverage in both sequences: 0
  Id: id1
    Length: 16
    Gaps: 0
    No coverage: 0
    Ambiguous: 0
  Id: id2
    Length: 16
    Gaps: 0
    No coverage: 6/16 (37.50%)
    No coverage locations (1-based): 6, 7, 8, 9, 10, 11
    Ambiguous: 0""",
            matchToString(match, read1, read2),
        )

    def testExcludeGapLocations(self):
        """
        If gap locations are not wanted, they should not appear in the result
        of a call to matchToString.
        """
        read1 = Read("id1", "TTTTTAAAAAAGCGCG")
        read2 = Read("id2", "TTTTT------GCGCG")
        match = compareDNAReads(read1, read2)
        self.assertEqual(
            """\
Exact matches: 10/16 (62.50%)
Ambiguous matches: 0
Mismatches: 6/16 (37.50%)
  Not involving gaps (i.e., conflicts): 0
  Involving a gap in one sequence: 6/16 (37.50%)
  Involving a gap in both sequences: 0
  Involving no coverage in one sequence: 0
  Involving no coverage in both sequences: 0
  Id: id1
    Length: 16
    Gaps: 0
    No coverage: 0
    Ambiguous: 0
  Id: id2
    Length: 16
    Gaps: 6/16 (37.50%)
    No coverage: 0
    Ambiguous: 0""",
            matchToString(match, read1, read2, includeGapLocations=False),
        )

    def testExcludeNoCoverageLocations(self):
        """
        If no coverage locations are not wanted, they should not appear in the
        result of a call to matchToString.
        """
        read1 = Read("id1", "TTTTTAAAAAAGCGCG")
        read2 = Read("id2", "TTTTT??????GCGCG")
        match = compareDNAReads(read1, read2, noCoverageChars="?")
        self.assertEqual(
            """\
Exact matches: 10/16 (62.50%)
Ambiguous matches: 0
Exact matches (ignoring no coverage sites): 10/10 (100.00%)
Ambiguous matches (ignoring no coverage sites): 0
Mismatches: 0
  Not involving gaps (i.e., conflicts): 0
  Involving a gap in one sequence: 0
  Involving a gap in both sequences: 0
  Involving no coverage in one sequence: 6/16 (37.50%)
  Involving no coverage in both sequences: 0
  Id: id1
    Length: 16
    Gaps: 0
    No coverage: 0
    Ambiguous: 0
  Id: id2
    Length: 16
    Gaps: 0
    No coverage: 6/16 (37.50%)
    Ambiguous: 0""",
            matchToString(match, read1, read2, includeNoCoverageLocations=False),
        )

    def testDifferenceCounts(self):
        """
        Difference counts must be returned correctly.
        """
        read1 = Read("id1", "GACTTAAAAAAGCGCG")
        read2 = Read("id2", "TTTTT------GCGCC")
        match = compareDNAReads(read1, read2)
        self.assertEqual(
            """\
Exact matches: 6/16 (37.50%)
Ambiguous matches: 0
Mismatches: 10/16 (62.50%)
  Not involving gaps (i.e., conflicts): 4/16 (25.00%)
  Involving a gap in one sequence: 6/16 (37.50%)
  Involving a gap in both sequences: 0
  Involving no coverage in one sequence: 0
  Involving no coverage in both sequences: 0
  Id: id1
    Length: 16
    Gaps: 0
    No coverage: 0
    Ambiguous: 0
  Id: id2
    Length: 16
    Gaps: 6/16 (37.50%)
    Gap locations (1-based): 6, 7, 8, 9, 10, 11
    No coverage: 0
    Ambiguous: 0
Difference counts:
    A->-: 6
    A->T: 1
    C->T: 1
    G->C: 1
    G->T: 1""",
            matchToString(match, read1, read2, includeDifferenceCounts=True),
        )

    def testDifferenceLocations(self):
        """
        Difference locations (and counts) must be returned correctly.
        """
        read1 = Read("id1", "GACTTAAAAAAGCGCG")
        read2 = Read("id2", "TTTTT------GCGCC")
        match = compareDNAReads(read1, read2)
        self.maxDiff = None
        self.assertEqual(
            """\
Exact matches: 6/16 (37.50%)
Ambiguous matches: 0
Mismatches: 10/16 (62.50%)
  Not involving gaps (i.e., conflicts): 4/16 (25.00%)
  Involving a gap in one sequence: 6/16 (37.50%)
  Involving a gap in both sequences: 0
  Involving no coverage in one sequence: 0
  Involving no coverage in both sequences: 0
  Id: id1
    Length: 16
    Gaps: 0
    No coverage: 0
    Ambiguous: 0
  Id: id2
    Length: 16
    Gaps: 6/16 (37.50%)
    Gap locations (1-based): 6, 7, 8, 9, 10, 11
    No coverage: 0
    Ambiguous: 0
Difference counts and locations:
    A->- (6): 6, 7, 8, 9, 10, 11
    A->T (1): 2
    C->T (1): 3
    G->C (1): 16
    G->T (1): 1""",
            matchToString(match, read1, read2, includeDifferenceLocations=True),
        )


class TestFindKozakConsensus(TestCase):
    """
    Test the findKozakConsensus function.
    """

    def testNoSequence(self):
        """
        If no sequence is given, no Kozak sequence should be found.
        """
        read = DNARead("id", "")
        self.assertEqual([], list(findKozakConsensus(read)))

    def testShortSequence(self):
        """
        If a 4 nt long sequence is given, no Kozak sequence should be found.
        """
        read = DNARead("id", "ATTG")
        self.assertEqual([], list(findKozakConsensus(read)))

    def testOneKozakConsensus(self):
        """
        In a given sequence with an exact Kozak consensus sequence, the offset
        and quality percentage should be as expected.
        """
        read = DNARead("id", "ATTGCCGCCATGGGGG")
        expectedKozakRead = DNAKozakRead(read, 3, 13, 100.0)
        (result,) = list(findKozakConsensus(read))
        self.assertEqual(expectedKozakRead, result)

    def testNoKozakConsensus(self):
        """
        In a given sequence without a Kozak consensus, the output should be
        as expected.
        """
        read = DNARead("id", "ATTGCCTCCATGGGGG")
        self.assertEqual([], list(findKozakConsensus(read)))

    def testFindTwoKozakConsensi(self):
        """
        In a given sequence with two Kozak consensuses with different offsets
        and qualities, the output should be as expected.
        """
        read = DNARead("id", "ATTGCCGCCATGGGGGGCCATGG")
        expectedRead1 = DNARead("id", "ATTGCCGCCATGGGGGGCCATGG")
        expectedRead2 = DNARead("id", "ATTGCCGCCATGGGGGGCCATGG")
        expectedKozakRead1 = DNAKozakRead(expectedRead1, 3, 13, 100.0)
        expectedKozakRead2 = DNAKozakRead(expectedRead2, 13, 23, 60.0)

        self.assertEqual(
            [expectedKozakRead1, expectedKozakRead2], list(findKozakConsensus(read))
        )

    def testNoKozakConsensusAtEnd(self):
        """
        In a given sequence without a Kozak consensus, the output should be
        as expected.
        """
        read = DNARead("id", "ATTGCCTCCATGGGGATG")
        self.assertEqual([], list(findKozakConsensus(read)))

    def testKozakConsensusAtEnd(self):
        """
        In a given sequence without a Kozak consensus, the output should be
        as expected.
        """
        read = DNARead("id", "AAAAAAATTGCCGCCATGG")
        expectedKozakRead = DNAKozakRead(read, 9, 19, 100.0)
        (result,) = list(findKozakConsensus(read))
        self.assertEqual(expectedKozakRead, result)

    def testKozakConsensusATGStart(self):
        """
        In a given sequence without a Kozak consensus, the output should be
        as expected.
        """
        read = DNARead("id", "AAAATGGAAAAAAATTGCCGCC")
        self.assertEqual([], list(findKozakConsensus(read)))

    def testKozakConsensusAtEndPart(self):
        """
        In a given sequence without a Kozak consensus, the output should be
        as expected.
        """
        read = DNARead("id", "AAAAAAATTGCCGCCATG")
        self.assertEqual([], list(findKozakConsensus(read)))


class TestFloatBaseCounts(TestCase):
    """
    Test the FloatBaseCounts class.
    """

    def testOneUnambiguousHomogeneous(self):
        """
        If one unambiguous code is passed to FloatBaseCounts, it must be
        considered homogeneous correctly, depending on the passed level.
        """
        counts = FloatBaseCounts("A")
        self.assertTrue(counts.homogeneous(1.0))
        self.assertTrue(counts.homogeneous(0.75))
        self.assertTrue(counts.homogeneous(0.0))

    def testTwoIdenticalUnambiguoussHomogeneous(self):
        """
        If an unambiguous code is passed to FloatBaseCounts twice, they must be
        considered homogeneous correctly, depending on the passed level.
        """
        counts = FloatBaseCounts("AA")
        self.assertTrue(counts.homogeneous(1.0))
        self.assertTrue(counts.homogeneous(0.75))
        self.assertTrue(counts.homogeneous(0.0))

    def testTwoDifferentUnambiguoussHomogeneous(self):
        """
        If two different unambiguous codes are passed to FloatBaseCounts,
        they must be considered homogeneous correctly, depending on the passed
        level.
        """
        counts = FloatBaseCounts("AG")
        self.assertFalse(counts.homogeneous(1.0))
        self.assertTrue(counts.homogeneous(0.5))
        self.assertTrue(counts.homogeneous(0.2))

    def testOneAmbiguousHomogeneousM(self):
        """
        If one ambiguous code (M = A, C) is passed to FloatBaseCounts, it must
        be considered homogeneous correctly, depending on the passed level.
        """
        counts = FloatBaseCounts("M")
        self.assertFalse(counts.homogeneous(1.0))
        self.assertFalse(counts.homogeneous(0.75))
        self.assertTrue(counts.homogeneous(0.0))

    def testOneAmbiguousHomogeneousV(self):
        """
        If one ambiguous code (V = A, C, G) is passed to FloatBaseCounts, it
        must be considered homogeneous correctly, depending on the passed
        level.
        """
        counts = FloatBaseCounts("V")
        self.assertFalse(counts.homogeneous(1.0))
        self.assertFalse(counts.homogeneous(0.75))
        self.assertTrue(counts.homogeneous(0.2))
        self.assertTrue(counts.homogeneous(0.0))

    def testTwoIdenticalUnambiguoussVariable(self):
        """
        If an unambiguous code is passed to FloatBaseCounts twice, they must be
        considered non-variable.
        """
        counts = FloatBaseCounts("AA")
        self.assertFalse(counts.variable())

    def testTwoDifferentUnambiguoussVariable(self):
        """
        If two different unambiguous codes are passed to FloatBaseCounts,
        they must be considered variable.
        """
        counts = FloatBaseCounts("AT")
        self.assertTrue(counts.variable())

    def testOneUnambiguousOneAmbiguousVariable(self):
        """
        If one unambiguous code and one incompatible ambiguous code are passed
        to FloatBaseCounts, they must be considered confirm variable.
        """
        counts = FloatBaseCounts("AS")
        self.assertTrue(counts.variable(confirm=True))

    def testOneUnambiguousOneAmbiguousNonVariable(self):
        """
        If one unambiguous code and one compatible ambiguous code are passed
        to FloatBaseCounts, they must not be considered confirm variable
        if confirm is True.
        """
        counts = FloatBaseCounts("AM")
        self.assertFalse(counts.variable(confirm=True))

    def testOneUnambiguousOneAmbiguousVariableUnconfirm(self):
        """
        If one unambiguous code and one compatible ambiguous code are passed
        to FloatBaseCounts, they must be considered variable if confirm
        is False.
        """
        counts = FloatBaseCounts("AM")
        self.assertTrue(counts.variable(confirm=False))

    def testOneUnambiguousOneAmbiguousVariableConfirm(self):
        """
        If one unambiguous code and one incompatible ambiguous code are passed
        to FloatBaseCounts, they must be considered variable if confirm
        is True.
        """
        counts = FloatBaseCounts("AY")
        self.assertTrue(counts.variable(confirm=True))

    def testTwoAmbiguousVariableConfirmFalse(self):
        """
        If two compatible but different ambiguous codes are passed to
        FloatBaseCounts, they must be considered variable if confirm
        is False.
        """
        counts = FloatBaseCounts("MR")
        self.assertTrue(counts.variable(confirm=False))

    def testTwoAmbiguousVariableConfirmTrue(self):
        """
        If two compatible but different ambiguous codes are passed to
        FloatBaseCounts, they must not be considered variable if
        confirm is True.
        """
        counts = FloatBaseCounts("MR")
        self.assertFalse(counts.variable(confirm=True))

    def testOneGapVariableUnconfirm(self):
        """
        If one gap and one unambiguous code are passed to FloatBaseCounts,
        they must be considered variable if confirm is False.
        """
        counts = FloatBaseCounts("A-")
        self.assertTrue(counts.variable(confirm=False))

    def testOneGapVariableConfirm(self):
        """
        If one gap and one unambiguous code are passed to FloatBaseCounts,
        they must be considered variable if confirm is True.
        """
        counts = FloatBaseCounts("A-")
        self.assertTrue(counts.variable(confirm=True))

    def testTwoAmbiguousStr(self):
        """
        If two compatible but different ambiguous codes are passed to
        FloatBaseCounts, they must be converted into a string correctly.
        """
        counts = FloatBaseCounts("MR")
        self.assertEqual("A:1.00 C:0.50 G:0.50 (0.500)", str(counts))

    def testTwoAmbiguousStrWithIntegerTotals(self):
        """
        If two 2-way ambiguous codes are passed to
        FloatBaseCounts, they must be converted into a string correctly
        (i.e., with integer counts).
        """
        counts = FloatBaseCounts("MMA")
        self.assertEqual("A:2 C:1 (0.667)", str(counts))

    def testLowerCase(self):
        """
        If two 2-way ambiguous codes are passed to FloatBaseCounts as lower
        case, they must be converted into a string correctly.
        """
        counts = FloatBaseCounts("mm")
        self.assertEqual("A:1 C:1 (0.500)", str(counts))

    def testMixedCase(self):
        """
        If two 2-way ambiguous codes are passed to FloatBaseCounts in mixed
        case, they must be converted into a string correctly.
        """
        counts = FloatBaseCounts("mM")
        self.assertEqual("A:1 C:1 (0.500)", str(counts))

    def testMostFrequentUnambiguous(self):
        """
        If one unambiguous code passed to FloatBaseCounts is most frequent, the
        mostFrequent method must give the expected result.
        """
        counts = FloatBaseCounts("AAACCTGG")
        self.assertEqual({"A"}, counts.mostFrequent())

    def testEquallyFrequent(self):
        """
        If two codes are passed to FloatBaseCounts in equal numbers, the
        mostFrequent method must give the expected result.
        """
        counts = FloatBaseCounts("AAACCCTGG")
        self.assertEqual(set("AC"), counts.mostFrequent())

    def testMostFrequentWithTwoAmbiguousStr(self):
        """
        If two overlapping ambiguous codes are passed to FloatBaseCounts, the
        mostFrequent method must give the expected result.
        """
        counts = FloatBaseCounts("MR")
        self.assertEqual({"A"}, counts.mostFrequent())

    def testHighestFrequencyUnambiguous(self):
        """
        If one unambiguous code passed to FloatBaseCounts is most frequent, the
        highestFrequency method must give the expected result.
        """
        counts = FloatBaseCounts("AAACCTGG")
        self.assertEqual(0.375, counts.highestFrequency())

    def testEqualFrequency(self):
        """
        If two codes are passed to FloatBaseCounts in equal numbers, the
        highestFrequency method must give the expected result.
        """
        counts = FloatBaseCounts("AAAACCCTGG")
        self.assertEqual(0.4, counts.highestFrequency())

    def testHighestFrequencyWithTwoAmbiguousStr(self):
        """
        If two overlapping ambiguous codes are passed to FloatBaseCounts, the
        highestFrequency method must give the expected result.
        """
        counts = FloatBaseCounts("MR")
        self.assertEqual(0.5, counts.highestFrequency())

    def testLength4Unambiguous(self):
        """
        If all unambiguous bases are given to FloatBaseCounts, its length
        must be 4.
        """
        counts = FloatBaseCounts("AAAACCCTGG")
        self.assertEqual(4, len(counts))

    def testLength3Ambiguous(self):
        """
        If overlapping ambiguous codes are passed to FloatBaseCounts, the
        length must give the expected result.
        """
        counts = FloatBaseCounts("MRC")
        self.assertEqual(3, len(counts))


class TestSequenceToRegex(TestCase):
    """
    Test the sequenceToRegex function.
    """

    def testEmpty(self):
        """
        The empty string should result in an empty regex.
        """
        self.assertEqual("", sequenceToRegex(""))

    def testUnambiguous(self):
        """
        An unambiguous string should result in an identical regex.
        """
        self.assertEqual("ACGT", sequenceToRegex("ACGT"))

    def testOneAmbiguous(self):
        """
        One ambiguous characters should result in the expected regex.
        """
        self.assertEqual("[AG]", sequenceToRegex("R"))

    def testTwoAmbiguous(self):
        """
        Two ambiguous characters should result in the expected regex.
        """
        self.assertEqual("[AG][ACG]", sequenceToRegex("RV"))

    def testMixed(self):
        """
        Mixed ambiguous and non-ambiguous characters should result in the
        expected regex.
        """
        self.assertEqual("A[AG]C[ACG]T", sequenceToRegex("ARCVT"))

    def testN(self):
        """
        An 'N' should result in an ACGT regex.
        """
        self.assertEqual("[ACGT]", sequenceToRegex("N"))

    def testQuestionMark(self):
        """
        A '?' should result in an ACGT regex.
        """
        self.assertEqual("[ACGT]", sequenceToRegex("?"))

    def testGap(self):
        """
        A '-' should result in an ACGT regex.
        """
        self.assertEqual("[ACGT]", sequenceToRegex("-"))

    def testWildcard(self):
        """
        An explicit wildcard should result in an ACGT regex.
        """
        self.assertEqual("[ACGT][ACGT][ACGT]", sequenceToRegex("*#!", wildcards="#*!"))

    def testUnknown(self):
        """
        An unknown character should result in a KeyError.
        """
        error = "^'5'$"
        self.assertRaisesRegex(KeyError, error, sequenceToRegex, "5")


class TestLeastAmbiguous(TestCase):
    """
    Test the leastAmbiguous function.
    """

    def testEmpty(self):
        """
        The empty string should result in a KeyError.
        """
        self.assertRaisesRegex(KeyError, "^''$", leastAmbiguous, "")

    def testUnknownNucleotides(self):
        """
        Unknown nucleotides should result in a KeyError.
        """
        self.assertRaisesRegex(KeyError, "^'123'$", leastAmbiguous, "123")

    def testDuplicationsIgnored(self):
        """
        If nucleotides are duplicated, there should be no problem.
        """
        self.assertEqual("A", leastAmbiguous("AAA"))

    def testDuplicationsDifferentCaseIgnored(self):
        """
        If nucleotides are duplicated in different cases, there should be no
        problem.
        """
        self.assertEqual("A", leastAmbiguous("AaaA"))

    def testSingleNucleotides(self):
        """
        Single nucleotides should be returned as themselves.
        """
        for base in "ACGT":
            self.assertEqual(base, leastAmbiguous(base))

    def testTwoNucleotides(self):
        """
        Two nucleotides should be handled correctly.
        """
        for bases, ambiguous in AMBIGUOUS_PAIRS:
            self.assertEqual(ambiguous, leastAmbiguous(bases))

    def testThreeNucleotides(self):
        """
        Three nucleotides should be handled correctly.
        """
        for bases, ambiguous in AMBIGUOUS_TRIPLES:
            self.assertEqual(ambiguous, leastAmbiguous(bases))

    def testFourNucleotides(self):
        """
        All four nucleotides should be handled correctly.
        """
        self.assertEqual("N", leastAmbiguous("AGCT"))

    def testFourNucleotidesOtherOrder(self):
        """
        All four nucleotides should be handled correctly when given in a
        different order.
        """
        self.assertEqual("N", leastAmbiguous("CGTA"))

    def testTuple(self):
        """
        All four nucleotides passed as a tuple should be handled correctly.
        """
        self.assertEqual("N", leastAmbiguous(tuple("AGCT")))

    def testList(self):
        """
        All four nucleotides passed as a list should be handled correctly.
        """
        self.assertEqual("N", leastAmbiguous(list("AGCT")))


class TestLeastAmbiguousFromBases(TestCase):
    """
    Test the leastAmbiguousFromCounts function.
    """

    def testNegativeCount(self):
        """
        A negative count must result in a ValueError.
        """
        self.assertRaisesRegex(
            ValueError,
            r"^Count for base 'A' is negative \(-1\)\.$",
            leastAmbiguousFromCounts,
            {"A": -1},
            0.9,
        )

    def testNegativeThreshold(self):
        """
        A negative threshold must result in a ValueError.
        """
        self.assertRaisesRegex(
            ValueError,
            r"^Threshold cannot be negative \(-0\.9\)\.$",
            leastAmbiguousFromCounts,
            {"A": 3},
            -0.9,
        )

    def testNoCounts(self):
        """
        If an empty dictionary of counts is passed, 'N' must result.
        """
        self.assertEqual("N", leastAmbiguousFromCounts({}, 0.9))

    def testAllCountsZero(self):
        """
        If the counts are all zero, 'N' must result.
        """
        self.assertEqual("N", leastAmbiguousFromCounts({"A": 0, "G": 0}, 0.9))

    def testOneBase(self):
        """
        If there is a single base, it must be returned.
        """
        for base in "ACGT":
            self.assertEqual(base, leastAmbiguousFromCounts({base: 3}, 0.9))

    def testTwoEqual(self):
        """
        If there are two bases with equal counts, the expected ambiguous code
        must be returned.
        """
        for bases, ambiguous in AMBIGUOUS_PAIRS:
            counts = dict.fromkeys(bases, 1)
            self.assertEqual(ambiguous, leastAmbiguousFromCounts(counts, 0.9))

    def testThreeEqual(self):
        """
        If there are three bases with equal counts, the expected ambiguous code
        must be returned.
        """
        for bases, ambiguous in AMBIGUOUS_TRIPLES:
            counts = dict.fromkeys(bases, 1)
            self.assertEqual(ambiguous, leastAmbiguousFromCounts(counts, 0.9))

    def testFourNucleotides(self):
        """
        If all four nucleotides have equal counts, 'N' must be returned.
        """
        counts = dict.fromkeys("AGCT", 1)
        self.assertEqual("N", leastAmbiguousFromCounts(counts, 0.9))

    def testOneOfTwoOverThreshold(self):
        """
        If one base of two is over the threshold, it must be returned.
        """
        for bases, _ in AMBIGUOUS_PAIRS:
            counts = {bases[0]: 3, bases[1]: 1}
            self.assertEqual(bases[0], leastAmbiguousFromCounts(counts, 0.5))

    def testOneOfTwoBelowThreshold(self):
        """
        If one base is dominant but not over the threshold, the ambiguous code
        must be returned.
        """
        for bases, ambiguous in AMBIGUOUS_PAIRS:
            counts = {bases[0]: 3, bases[1]: 1}
            self.assertEqual(ambiguous, leastAmbiguousFromCounts(counts, 0.9))

    def testOneOfThreeOverThreshold(self):
        """
        If one base of three is over the threshold, it must be returned.
        """
        for bases, _ in AMBIGUOUS_TRIPLES:
            counts = {bases[0]: 3, bases[1]: 1, bases[2]: 1}
            self.assertEqual(bases[0], leastAmbiguousFromCounts(counts, 0.5))

    def testTwoOfThreeOverThreshold(self):
        """
        If two bases of three are over the threshold, the code for those two
        must be returned.
        """
        for bases, _ in AMBIGUOUS_TRIPLES:
            counts = {bases[0]: 4, bases[1]: 4, bases[2]: 2}
            self.assertEqual(
                leastAmbiguous(bases[:2]), leastAmbiguousFromCounts(counts, 0.8)
            )

    def testGeneiousExamplesNoTie(self):
        """
        Test the no-tied counts example from
        https://assets.geneious.com/manual/2020.1/static/GeneiousManualse43.html
        """
        counts = {"A": 6, "G": 3, "T": 1}
        self.assertEqual("A", leastAmbiguousFromCounts(counts, 0.4))
        self.assertEqual("R", leastAmbiguousFromCounts(counts, 0.7))
        self.assertEqual("D", leastAmbiguousFromCounts(counts, 0.95))

    def testGeneiousExamplesTie(self):
        """
        Test the tied counts example from
        https://assets.geneious.com/manual/2020.1/static/GeneiousManualse43.html
        """
        counts = {"A": 6, "G": 2, "T": 2}
        self.assertEqual("A", leastAmbiguousFromCounts(counts, 0.4))
        self.assertEqual("D", leastAmbiguousFromCounts(counts, 0.7))
        self.assertEqual("D", leastAmbiguousFromCounts(counts, 0.95))


class TestBases(TestCase):
    """
    Test the Bases class.
    """

    def testEmptyIsFalse(self):
        """
        An empty Bases instance must be considered False.
        """
        self.assertFalse(Bases())

    def testEmptyHasLengthZero(self):
        """
        An empty Bases instance must have length zero.
        """
        self.assertEqual(0, len(Bases()))

    def testLengthOneIsTrue(self):
        """
        If we append a (base, quality) pair, the instance must be considered
        True.
        """
        b = Bases().append("G", 30)
        self.assertTrue(b)

    def testLengthOne(self):
        """
        If we append a (base, quality) pair, the length must be one.
        """
        b = Bases().append("G", 30)
        self.assertEqual(1, len(b))

    def testGetitem(self):
        """
        __getitem__ must work as expected.
        """
        b = Bases().append("G", 30)
        self.assertEqual(30, b["G"])

    def testAdd(self):
        """
        Addition must work as expected.
        """
        b1 = Bases().append("G", 30)
        b2 = Bases().append("G", 20)
        b3 = b1 + b2

        self.assertEqual(1, len(b1))
        self.assertEqual(1, len(b2))
        self.assertEqual(2, len(b3))
        self.assertEqual(50, b3.counts["G"])

    def testIAdd(self):
        """
        In-place addition must work as expected.
        """
        b1 = Bases().append("G", 30)
        b2 = Bases().append("G", 20)
        b2 += b1

        self.assertEqual(1, len(b1))
        self.assertEqual(2, len(b2))
        self.assertEqual(50, b2.counts["G"])

    def testConsensusNoReads(self):
        """
        The consensus method must return the no coverage string if there are
        no reads.
        """
        b = Bases()
        self.assertEqual("Z", b.consensus(0.8, 1, "L", "Z"))

    def testConsensusLowReads(self):
        """
        The consensus method must return the low coverage string if there are
        no reads.
        """
        b = Bases().append("G", 20)
        self.assertEqual("L", b.consensus(0.8, 2, "L", "Z"))

    def testConsensusOneBase(self):
        """
        If there is a single base, it must be returned.
        """
        for base in "ACGT":
            b = Bases().append(base, 20)
            self.assertEqual(base, b.consensus(0.8, 1, "L", "Z"))

    def testConsensusTwoEqual(self):
        """
        If there are two bases with equal counts, the expected ambiguous code
        must be returned.
        """
        for bases, ambiguous in AMBIGUOUS_PAIRS:
            b = Bases().append(bases[0], 20).append(bases[1], 20)
            self.assertEqual(ambiguous, b.consensus(0.8, 1, "L", "Z"))

    def testConsensusThreeEqual(self):
        """
        If there are three bases with equal counts, the expected ambiguous code
        must be returned.
        """
        for bases, ambiguous in AMBIGUOUS_TRIPLES:
            b = Bases().append(bases[0], 20).append(bases[1], 20).append(bases[2], 20)
            self.assertEqual(ambiguous, b.consensus(0.8, 1, "L", "Z"))

    def testConsensusFourNucleotides(self):
        """
        If all four nucleotides have equal counts, 'N' must be returned.
        """
        b = Bases()
        for base in "ACGT":
            b.append(base, 20)
        self.assertEqual("N", b.consensus(0.8, 1, "L", "Z"))



#!/usr/bin/env python

import sys

from Bio.Seq import Seq

if len(sys.argv) == 0:
    print(f"Usage: {sys.argv[0]} DNA-sequence", file=sys.stderr)
    sys.exit(1)

for dna in sys.argv[1:]:
    print(Seq(dna).translate())



#!/usr/bin/env python

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=("Given FASTA on stdin, print the number of sequences to stdout."),
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    reads = parseFASTACommandLineOptions(args)

    count = 0
    for read in reads:
        count += 1

    print(count)



from io import StringIO
from subprocess import CalledProcessError
from unittest import TestCase

from dark.process import Executor


class TestProcess(TestCase):
    """
    Test the Process class.
    """

    def testUnknownCommand(self):
        """
        An unknown command must raise CalledProcessError.
        """
        e = Executor()
        # Presumably there will not be an executable with this name!
        command = "/".join(["dev", "non-existent", "@" * 20])
        error = rf"^Command '{command}' returned non-zero exit status 127\.$"
        self.assertRaisesRegex(CalledProcessError, error, e.execute, command)

    def testDryRunTrue(self):
        """
        The dryRun attribute must be set when dryRun=True.
        """
        e = Executor(dryRun=True)
        self.assertTrue(e.dryRun)

    def testDryRunFalse(self):
        """
        The dryRun attribute must be set when dryRun=False.
        """
        e = Executor(dryRun=False)
        self.assertFalse(e.dryRun)

    def testDryRunDefault(self):
        e = Executor(dryRun=True)
        result = e.execute("date")
        self.assertIsNone(result)
        self.assertEqual("$ date", e.log[-1])

    def testDryRunDefaultOverride(self):
        """
        It must be possible to override the default dryRun setting by passing
        a value to C{execute}.
        """
        e = Executor(dryRun=False)
        result = e.execute("date", dryRun=True)
        self.assertIsNone(result)
        self.assertEqual("$ date", e.log[-1])

    def testEchoStr(self):
        """
        We should be able to call echo using a string command.
        """
        e = Executor()
        result = e.execute("echo hello")
        self.assertEqual("hello", result.stdout.strip())
        self.assertTrue("$ echo hello" in e.log)

    def testEchoList(self):
        """
        We should be able to call echo using a list command.
        """
        e = Executor()
        result = e.execute(["echo", "hello"])
        self.assertEqual("hello", result.stdout.strip())
        self.assertTrue("$ echo hello" in e.log)

    def testPipe(self):
        """
        We should be able to pipe echo into wc -c.
        """
        e = Executor()
        result = e.execute("echo hello | wc -c")
        self.assertEqual("6", result.stdout.strip())
        self.assertTrue("$ echo hello | wc -c" in e.log)

    def testPassStdoutFileDescriptorDryRun(self):
        """
        Passing a file descriptor for stdout when in dry run mode must result
        in the command being written to that descriptor.
        """
        stdout = StringIO()
        e = Executor(stdout=stdout, dryRun=True)
        e.execute("echo hello")
        self.assertEqual("$ echo hello\n", stdout.getvalue())

    def testPassStdoutFileDescriptor(self):
        """
        Passing a file descriptor for stdout must result in the
        command and its output being written to that descriptor.
        """
        stdout = StringIO()
        e = Executor(stdout=stdout)
        e.execute("echo hello")
        self.assertEqual("$ echo hello\nhello\n", stdout.getvalue())

    def testPassStdoutFileDescriptorToExecute(self):
        """
        Passing a file descriptor for stdout in the call to execute must
        result in the command and its output being written to that descriptor.
        """
        stdout = StringIO()
        e = Executor()
        e.execute("echo hello", stdout=stdout)
        self.assertEqual("$ echo hello\nhello\n", stdout.getvalue())

    def testPassNoneStdoutFileDescriptorToExecute(self):
        """
        Passing None for stdout in the call to execute must result in the
        command and its output NOT being written to the corresponding option
        passed to the class __init__ (i.e., it must be possible to override
        the default for the class instance).
        """
        stdout = StringIO()
        e = Executor(stdout=stdout)
        e.execute("echo hello", stdout=None)
        self.assertEqual("", stdout.getvalue())

    def testPassStderrFileDescriptor(self):
        """
        Passing a file descriptor for stderr must result in the
        standard error output being written to that descriptor.
        """
        stderr = StringIO()
        e = Executor(stderr=stderr)
        e.execute("echo hello >&2")
        self.assertEqual("hello\n", stderr.getvalue())

    def testPassStderrFileDescriptorToExecute(self):
        """
        Passing a file descriptor for stderr in the call to execute must
        result in the standard error output being written to that descriptor.
        """
        stderr = StringIO()
        e = Executor()
        e.execute("echo hello >&2", stderr=stderr)
        self.assertEqual("hello\n", stderr.getvalue())

    def testPassStdoutAndStderrFileDescriptors(self):
        """
        Passing file descriptors for stdout and stderr must result in the
        command and the standard output being written to the former and the
        standard error being written to the latter.
        """
        stdout = StringIO()
        stderr = StringIO()
        e = Executor(stdout=stdout, stderr=stderr)
        e.execute("echo one; echo two >&2")
        self.assertEqual("$ echo one; echo two >&2\none\n", stdout.getvalue())
        self.assertEqual("two\n", stderr.getvalue())



import numpy as np


class BaseImage:
    """
    Hold a representation of colored sequence bases in a 2D grid suitable
    for placing onto an alignment graph (see utils.py).
    """

    def __init__(self, xRange: int, yRange: int, xScale: int = 1, yScale: int = 1):
        # np.ones is passed a (y range, x range) dimension.
        self.data = np.ones(
            (yRange * yScale + 1, xRange * xScale + 1), dtype=(float, 3)
        )
        self.xScale = xScale
        self.yScale = yScale

    def set(self, x: int, y: int, value: float):
        """
        Set the data at (x, y) to value.
        """
        xBase = int(x) * self.xScale
        yBase = int(y) * self.yScale
        for xOffset in range(self.xScale):
            for yOffset in range(self.yScale):
                self.data[yBase + yOffset, xBase + xOffset] = value



#!/usr/bin/env python

import argparse

from dark.trees import removeAlrt


def makeParser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            """Read a Newick tree file produced by iqtree2 with both alrt
and support values on nodes, drop the alrt values and optionally
adjust the support values (based on the passed bootstrap support
threshold and alrt threshold). Print the new tree as a (Newick) string."""
        ),
    )

    parser.add_argument(
        "treefile",
        help="The Newick tree file to read.",
    )

    parser.add_argument(
        "--supportThreshold",
        default=95,
        type=int,
        help="The integer minimum UF boot support value.",
    )

    parser.add_argument(
        "--alrtThreshold",
        default=80.0,
        type=float,
        help="The float minimum alrt value.",
    )

    parser.add_argument(
        "--adjustSupportValues",
        action="store_true",
        help=(
            "The support on branches that have support values at least "
            "--supportThreshold but alrt values below --alrtThreshold "
            "will be set to the --replacementSupportValue value."
        ),
    )

    parser.add_argument(
        "--replacementSupportValue",
        help=(
            "The string value to use as a label on branches whose support "
            "values are adjusted."
        ),
    )

    parser.add_argument(
        "--expectedUnlabeledNodeCount",
        type=int,
        help="The number of unlabeled nodes (in the input file) to allow.",
    )

    parser.add_argument(
        "--verbose",
        action="store_true",
        help=(
            "Print information about how many branches have good support "
            "but too-low alrt values."
        ),
    )

    return parser


def main() -> None:
    args = makeParser().parse_args()

    print(
        removeAlrt(
            args.treefile,
            args.supportThreshold,
            args.alrtThreshold,
            args.adjustSupportValues,
            args.replacementSupportValue,
            args.expectedUnlabeledNodeCount,
            args.verbose,
        )
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import sys
import tempfile

from dark.reads import addFASTACommandLineOptions, parseFASTACommandLineOptions

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "Given FASTA sequences (of equal length) on stdin write Phylip to stdout."
        )
    )

    parser.add_argument(
        "--addIdSpace",
        action="store_true",
        default=False,
        help=(
            "If True, print an extra space after each sequence id (the "
            "space is expected by some programs that process Phylip files, "
            "e.g., baseml)."
        ),
    )

    addFASTACommandLineOptions(parser)
    args = parser.parse_args()
    idSpace = " " if args.addIdSpace else ""
    reads = parseFASTACommandLineOptions(args)

    count = 0
    first = True
    length: int
    with tempfile.TemporaryFile() as fp:
        for read in reads:
            count += 1
            if first:
                first = False
                length = len(read)
            else:
                if len(read) != length:
                    raise ValueError(
                        "FASTA sequence %r was not of the "
                        "expected length (%d)" % (read.id, length)
                    )

            name = read.id.split()[0]
            fp.write(("%s%s %s\n" % (name, idSpace, read.sequence)).encode("utf-8"))

        if count:
            # Print the phylip header with the number of sequences and their
            # common length.
            print("%d %d" % (count, length))
            fp.seek(0)
            write = sys.stdout.write
            while True:
                chunk = fp.read(1024)
                if chunk:
                    write(chunk.decode("utf-8"))
                else:
                    break



import os
from contextlib import contextmanager

import progressbar  # type: ignore


@contextmanager
def maybeProgressBar(show, maxValue, prefix):
    """
    A context manager to maybe show a progress bar.

    @param show: If C{True}, yield a progress bar, else a Class with an
        C{update} method that does nothing.
    @param maxValue: The C{int} number of tasks to show progress for.
    @param prefix: A C{str} prefix, to appear at the start of the progress bar.
    """
    if show and os.isatty(2):
        widgets = [
            progressbar.SimpleProgress(format="%(value_s)s/%(max_value_s)s"),
            progressbar.Percentage(format=" %(percentage)3d%%"),
            " ",
            progressbar.Bar(marker="\x1b[33m#\x1b[39m"),
            " ",
            progressbar.Timer(format="Elapsed: %(elapsed)s"),
            " ",
            progressbar.ETA(format="ETA: %(eta)8s"),
        ]
        with progressbar.ProgressBar(
            max_value=maxValue, widgets=widgets, prefix=prefix
        ) as bar:
            yield bar
    else:

        class Bar:
            update = staticmethod(lambda _: None)

        yield Bar



#!/usr/bin/env python

from itertools import count
from math import log10
from typing import Iterator, Optional

from dark.dimension import dimensionalIterator


def _commonPrefix(a: str, b: str) -> str:
    prefix = []
    for i, j in zip(a, b):
        if i == j:
            prefix.append(i)
        else:
            break

    return "".join(prefix)


def _doInts(
    start: int, end: Optional[int], zeroes: bool, unlimited: bool
) -> Iterator[str]:
    """
    Yield integer ids.
    """
    width = "0" + str(int(log10(end)) + 1) if (zeroes and end is not None) else ""

    if end is None:
        if unlimited:
            # Just produce numbers from start up to (but not including) the
            # next power of ten.
            it = iter(range(start, 10 ** (int(log10(start)) + 1)))
        else:
            # A limit imposed (by the use of maxResults in the call to ids), so it's
            # safe to use itertools.count.
            it = iter(count(start))
    else:
        if start > end:
            start, end = end, start
        it = iter(range(start, end + 1))

    for i in it:
        yield f"{i:{width}d}"


def _doStrings(start: str, end: Optional[str]) -> Iterator[str]:
    if end is not None and len(start) != len(end):
        raise ValueError("Range specifiers must be of equal length.")

    bases = []
    dimensions = []
    odometerStart = []
    parts: list[Optional[str]] = []

    for letter in start:
        if letter.islower():
            first = ord("a")
            bases.append(first)
            dimensions.append(26)
            parts.append(None)
        elif letter.isupper():
            first = ord("A")
            bases.append(first)
            dimensions.append(26)
            parts.append(None)
        elif letter.isdigit():
            first = ord("0")
            bases.append(first)
            dimensions.append(10)
            parts.append(None)
        else:
            # Save this as a constant/unchanging letter seeing as it's not
            # clear how to iterate it.
            parts.append(letter)
            continue

        odometerStart.append(ord(letter) - first)

    for odometer in dimensionalIterator(dimensions, start=odometerStart):
        variable = [f"{chr(bases[i] + c)}" for i, c in enumerate(odometer)]
        resultParts = []
        variableIndex = 0
        for part in parts:
            if part is None:
                resultParts.append(variable[variableIndex])
                variableIndex += 1
            else:
                resultParts.append(part)

        # Make sure we consumed all variable parts from the dimensional iterator.
        assert variableIndex == len(variable)

        result = "".join(resultParts)

        yield result

        if end is not None and result == end:
            break


def ids(
    start: str | int,
    end: Optional[str | int] = None,
    prefix: Optional[str] = None,
    sep: str = " ",
    zeroes: bool = False,
    maxResults: Optional[int] = None,
) -> Iterator[str]:
    """
    Yield ids, starting from C{start}.
    """
    if end is None:
        prefix = prefix or ""
        rest1 = str(start)
        rest2 = None
    else:
        start, end = str(start), str(end)
        if prefix is None:
            prefix = _commonPrefix(start, end)
            rest1 = start[len(prefix) :]
            rest2 = end[len(prefix) :]
        else:
            rest1, rest2 = start, end

    ints = True
    try:
        int1 = int(rest1)
    except ValueError:
        ints = False
    else:
        if rest2 is None:
            int2 = None
        else:
            try:
                int2 = int(rest2)
            except ValueError:
                ints = False

    suffixes = (
        _doInts(int1, int2, zeroes, maxResults is None)
        if ints
        else _doStrings(rest1, rest2)
    )

    resultCount = 0
    for suffix in suffixes:
        if maxResults is not None and resultCount >= maxResults:
            return

        yield f"{prefix}{suffix}"
        resultCount += 1



import bz2
import gzip
import itertools
import os
import re
import string
from contextlib import contextmanager
from math import log
from os.path import basename
from pathlib import Path
from re import compile
from statistics import median as _median
from time import gmtime, strftime
from typing import Iterable, List, Optional

from sklearn.metrics.cluster import entropy

from dark import File


def numericallySortFilenames(names: List[str]) -> List[str]:
    """
    Sort (ascending) a list of file names by their numerical prefixes.
    The number sorted on is the numeric prefix of the basename of
    the given filename. E.g., '../output/1.json.bz2' will sort before
    '../output/10.json.bz2'.

    @param: A C{list} of file names, each of whose basename starts with a
        string of digits.
    @return: The sorted C{list} of full file names.
    """

    def numericPrefix(name: str) -> int:
        """
        Find any numeric prefix of C{name} and return it as an C{int}.

        @param: A C{str} file name, whose name possibly starts with digits.
        @return: The C{int} number at the start of the name, else 0 if
            there are no leading digits in the name.
        """
        count = 0
        for ch in name:
            if ch in string.digits:
                count += 1
            else:
                break
        return 0 if count == 0 else int(name[0:count])

    return sorted(names, key=lambda name: numericPrefix(basename(name)))


def median(numbers):
    """
    Find the median of a set of numbers.

    @param numbers: A C{list} of numeric values.
    @raise ValueError: If C{l} is empty.
    @return: The median of C{numbers}.
    """
    if len(numbers) == 0:
        # Match the empty-argument exception message of Python's built-in max
        # and min functions.
        raise ValueError("arg is an empty sequence")
    else:
        return _median(numbers)


@contextmanager
def asHandle(fileNameOrHandle: File, mode="rt", encoding="UTF-8"):
    """
    Decorator for file opening that makes it easy to open compressed files
    and which can be passed an already-open file handle or a file name.
    Based on L{Bio.File.as_handle}.

    @param fileNameOrHandle: Either a C{str} or a file handle.
    @param mode: The C{str} mode to use for opening the file.
    @param encoding: The C{str} encoding to use when opening the file.
    @return: A generator that can be turned into a context manager via
        L{contextlib.contextmanager}.
    """
    if isinstance(fileNameOrHandle, (Path, str)):
        fileNameOrHandle = str(fileNameOrHandle)
        if "b" in mode:
            # Binary modes don't take encodings.
            encoding = None
        if fileNameOrHandle.endswith(".gz"):
            yield gzip.open(fileNameOrHandle, mode=mode, encoding=encoding)
        elif fileNameOrHandle.endswith(".bz2"):
            yield bz2.open(fileNameOrHandle, mode=mode, encoding=encoding)
        else:
            with open(fileNameOrHandle, mode=mode, encoding=encoding) as fp:
                yield fp
    else:
        yield fileNameOrHandle


_rangeRegex = compile(r"^\s*(\d+)(?:\s*-\s*(\d+))?\s*$")


# Note: the parseRangeExpression, which uses the following function as a
#       helper (see below) is more general / powerful than this function on
#       its own.
def parseRangeString(s, convertToZeroBased=False, asList: bool = False):
    """
    Parse a range string of the form 1-5,12,100-200.

    @param s: A C{str} specifiying a set of numbers, given in the form of
        comma separated numeric ranges or individual indices.
    @param convertToZeroBased: If C{True} all indices will have one
        subtracted from them.
    @raise ValueError: If the range in C{s} cannot be parsed.
    @return: A C{set} of all C{int}s in the specified set.
    """

    result = []
    seen = set()
    for _range in s.split(","):
        match = _rangeRegex.match(_range)
        if match:
            start, end = match.groups()
            start = int(start)
            end = start if end is None else int(end)
            if start > end:
                start, end = end, start

            toAdd = (
                range(start - 1, end) if convertToZeroBased else range(start, end + 1)
            )

            for n in toAdd:
                if n not in seen:
                    seen.add(n)
                    result.append(n)
        else:
            raise ValueError(
                "Illegal range %r. Ranges must single numbers or "
                "number-number." % _range
            )

    return result if asList else set(result)


# The following matches range expressions such as
#
#   3
#   3-4
#   3-4,5
#   3-4,6,9-10,99
#
# including any embedded whitespace.
_rangeExpressionRegex = re.compile(r"\d+(:?\s*-\s*\d+)?(:?\s*,\s*\d+(:?\s*-\s*\d+)?)*")


def parseRangeExpression(s, convertToZeroBased=False):
    """
    Parse a range string expression of the form "1-5,6 & (12 | 100-200)".

    @param s: A C{str} specifiying a range expression, given in the form of
        comma separated numeric ranges or individual indices, interspersed
        with (), |, and &.
    @param convertToZeroBased: If C{True} all indices will have one
        subtracted from them.
    @raise ValueError: If the expression in C{s} cannot be evaluated.
    @return: A C{set} of all C{int}s in the specified expression.
    """
    if not s:
        return set()

    pos = 0
    expr = ""
    for match in _rangeExpressionRegex.finditer(s):
        start, end = match.span()
        if start > pos:
            expr += s[pos:start]
        expr += "{%s}" % ",".join(
            map(str, parseRangeString(match.group(0), convertToZeroBased))
        )
        pos = end

    if pos < len(s):
        expr += s[pos:]

    try:
        return eval(expr)
    except Exception:
        raise ValueError(expr)


def baseCountsToStr(counts: dict[str, int]) -> str:
    """
    Convert base counts to a string.

    @param counts: A C{Counter} instance.
    @return: A C{str} representation of nucleotide counts at an offset.
    """
    return " ".join([("%s:%d" % (base, counts[base])) for base in sorted(counts)])


def nucleotidesToStr(nucleotides, prefix=""):
    """
    Convert offsets and base counts to a string.

    @param nucleotides: A C{defaultdict(Counter)} instance, keyed
        by C{int} offset, with nucleotides keying the Counters.
    @param prefix: A C{str} to put at the start of each line.
    @return: A C{str} representation of the offsets and nucleotide
        counts for each.
    """
    result = []
    for offset in sorted(nucleotides):
        result.append(
            "%s%d: %s" % (prefix, offset, baseCountsToStr(nucleotides[offset]))
        )
    return "\n".join(result)


def countPrint(mesg: str, count: int, len1: int, len2: Optional[int] = None) -> str:
    """
    Format a message followed by an integer count and a percentage (or
    two, if the sequence lengths are unequal).

    @param mesg: a C{str} message.
    @param count: a numeric value.
    @param len1: the C{int} length of sequence 1.
    @param len2: the C{int} length of sequence 2. If C{None}, will
        default to C{len1}.
    @return: A C{str} for printing.
    """

    def percentage(a: int, b: int) -> float:
        """
        What percent of a is b?

        @param a: a numeric value.
        @param b: a numeric value.
        @return: the C{float} percentage.
        """
        return 100.0 * a / b if b else 0.0

    if count == 0:
        return "%s: %d" % (mesg, count)
    else:
        len2 = len2 or len1
        if len1 == len2:
            return "%s: %d/%d (%.2f%%)" % (mesg, count, len1, percentage(count, len1))
        else:
            return "%s: %d/%d (%.2f%%) of sequence 1, %d/%d (%.2f%%) of sequence 2" % (
                mesg,
                count,
                len1,
                percentage(count, len1),
                count,
                len2,
                percentage(count, len2),
            )


def pct(
    a: int, b: int, commas: bool = False, places: int = 3, denom: bool = True
) -> str:
    """
    Format a string showing two integers and what percentage the first
    is of the second.

    @param a: An C{int}, the numerator.
    @param b: An C{int}, the denominator.
    @param commas: Include thousand-separating commas in the result.
    @param places: The C{int} number of decimal places to include in the fraction.
    @param denom: Include the denominator.
    """
    assert 0 <= a <= b
    frac = a / b * 100.0 if b else 0.0
    numerator = f"{a:,}" if commas else str(a)
    denominator = (f"/{b:,}" if commas else f"/{b}") if denom else ""
    return f"{numerator}{denominator} ({frac:.{places}f}%)"


@contextmanager
def cd(newdir):
    """
    Trivial context manager for temporarily switching directory.

    @param dir: A C{str} directory to cd to.
    """
    olddir = os.getcwd()
    try:
        os.chdir(newdir)
        yield newdir
    finally:
        os.chdir(olddir)


def take(iterator, n):
    """
    Repeatedly yield lists of C{n} items taken from an iterator.

    @param iterator: Something that can be iterated.
    @param n: An C{int} number of items to yield each time, greater than zero.
    @raise AssertionError: If C{n} < 1.
    @return: A generator that yields lists of length C{n}.
    """
    assert n > 0
    items = []
    iterator = iter(iterator)

    while True:
        try:
            item = next(iterator)
        except StopIteration:
            if items:
                yield items
            break
        else:
            items.append(item)
            if len(items) == n:
                yield items
                items = []


# This function should have a more general name.
def readLabels(fp):
    """
    Read a file of replacement label names.

    @param fp: An open file pointer to read. The file must contain lines with
        a name, a TAB, then a replacement name.
    @return: A C{dict} mapping old names to new names.
    """
    result = {}
    for line in fp:
        oldName, newName = map(str.strip, line.split("\t"))
        result[oldName] = newName
    return result


def matchOffset(leftPaddedReference, leftPaddedQuery):
    """
    At what reference offset does a query begin to match a reference?

    @param leftPaddedReference: A left-padded reference C{str}.
    @param leftPaddedQuery: A left-padded query C{str}.
    @return: An C{int} offset into the (unpadded) reference.
    """
    offset = 0
    for queryChar, referenceChar in zip(leftPaddedQuery, leftPaddedReference):
        if queryChar not in " -":
            break
        offset += referenceChar not in " -"

    return offset


@contextmanager
def openOr(filename, mode="r", defaultFp=None, specialCaseHyphen=True):
    """
    A context manager to either open a file or use a pre-opened default file.

    @param filename: If not C{None}, this is the argument to pass to C{open}
        along with C{mode}. If C{None}, C{fp} is used.
    @param mode: The C{str} file opening mode, used if C{filename} is not
        C{None}.
    @param defaultFp: An open file-like object to yield if C{filename} is
        C{None}.
    @param specialCaseHyphen: If C{True}, treat '-' as a C{None} filename
        and yield the C{defaultFp}.
    """
    if filename is None or filename == "-" and specialCaseHyphen:
        yield defaultFp
    else:
        with open(filename, mode) as fp:
            yield fp


def intsToIntervals(numbers: Iterable[int]) -> list[tuple[int, int]]:
    """
    Produce a list of consecutive integer intervals from a set of integers.

    Given a set of integers, A, this function finds a mutually disjoint set of intervals
    [a_i, b_i] (for i in some finite index set), such that A is the union of [a_i, b_i]
    over all i and returns a list of ordered-pairs (a_i, b_i).

    @param numbers: An iterable of integer values.
    @return: A C{list} of C{int} (start, end) consecutive intervals, sorted from least
        to greatest. Each tuple holds a pair of (inclusive) numbers that are
        consecutive in C{numbers}. In each pair, C{start} will equal C{end} if
        C{start}+1 is not present in numbers.
    """
    remaining = set(numbers)
    assert all(isinstance(x, int) for x in remaining)
    intervals: list[tuple[int, int]] = []

    while remaining:
        start = min(remaining)
        remaining.remove(start)

        for end in itertools.count(start + 1):
            if end in remaining:
                remaining.remove(end)
            else:
                end -= 1
                break

        intervals.append((start, end))

    return intervals


def intsToStringIntervals(
    numbers: Iterable[int], sep: str = "-", minimize: bool = False
) -> list[str]:
    """
    Make a list of strings of consecutive intervals given a set of integers.

    @param numbers: An iterable of integer values.
    @param sep: The C{str} separator to put between the start and end integers
        when representing a consecutive range of numbers in C{numbers}.
    @param minimize: If true, make a further simplification, changing an interval
        such as 2016-2020 into 2016-20 or 2018-2019 into 2018-9.
    @return: A C{list} of C{str} intervals, sorted from least to greatest.
    """
    intervals = intsToIntervals(numbers)

    if minimize:
        minimized = []
        for start, end in intervals:
            if start == end:
                minimized.append(str(start))
            else:
                oneIsNegative = start < 0 and end >= 0
                sstart, send = map(str, (start, end))
                if len(sstart) == len(send) and not oneIsNegative:
                    for commonPrefixLen in range(len(sstart)):
                        if sstart[commonPrefixLen] != send[commonPrefixLen]:
                            break
                    minimized.append(sstart + sep + send[commonPrefixLen:])
                else:
                    minimized.append(sstart + sep + send)

        return minimized

    else:
        return [
            (str(start) if start == end else str(start) + sep + str(end))
            for (start, end) in intervals
        ]


def gmt() -> str:
    return strftime("%Y-%m-%d %H:%M:%S (UTC)", gmtime())


_LOG2 = log(2.0)


def entropy2(labels: Iterable[str]) -> float:
    return entropy(labels) / _LOG2



#!/usr/bin/env python

"""
Print a (Levenshtein) distance matrix for a set of known adaptors
given on the command line.

For an example of usage,
see https://notebooks.antigenic-cartography.org/terry/emc-adaptors.html
"""

from dark.distance import levenshtein

if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser(
        description=(
            "Print a (Levenshtein) distance matrix for a set of known adaptors"
        )
    )

    parser.add_argument(
        "adaptors",
        nargs="+",
        metavar="adaptor",
        help="The set of adaptors used in sequencing.",
    )

    args = parser.parse_args()

    adaptors = args.adaptors
    nAdaptors = len(adaptors)
    length = len(adaptors[0])
    spaces = " " * length

    for i in range(length):
        print(spaces, end=" ")
        for adaptor in adaptors:
            print(adaptor[i], end=" ")
        print()

    for i in range(nAdaptors):
        print(adaptors[i], end=" ")
        for j in range(nAdaptors):
            if j < i:
                print(" ", end=" ")
            else:
                print(levenshtein(adaptors[i], adaptors[j]), end=" ")
        print()



from pysam import CHARD_CLIP, CINS, CSOFT_CLIP  # type: ignore

from dark.sam import CONSUMES_REFERENCE

# From https://samtools.github.io/hts-specs/SAMv1.pdf
(CINS_STR, CDEL_STR, CMATCH_STR, CEQUAL_STR, CDIFF_STR, CHARD_CLIP_STR) = tuple(
    "IDM=XH"
)


def dna2cigar(s1, s2, concise=False):
    """
    Form a CIGAR string from two equal-length DNA sequences. Case is not
    ignored, so if you want to ignore case you must convert the sequences
    to a consistent case (upper or lower, doesn't matter) before calling.

    @param s1: A C{str} of DNA 'A', 'G', 'C', 'T' characters.
    @param s2: A C{str} of DNA 'A', 'G', 'C', 'T' characters.
    @param concise: If C{True}, use 'M' for matches and mismatches instead
        of the more specific 'X' and '='.
    @raise ValueError: If the two sequences are not of equal length or if the
        sequences are of length zero.
    @return: A C{str} CIGAR string.
    """
    len1 = len(s1)
    if len1 != len(s2):
        raise ValueError(
            "Sequences %r and %r of unequal length (%d != %d)."
            % (s1, s2, len1, len(s2))
        )

    if len1 == 0:
        raise ValueError("Two sequences of zero length were passed.")

    if concise:
        return "%d%s" % (len1, CMATCH_STR)

    result = []
    length, operation = 0, None

    for n1, n2 in zip(s1, s2):
        if n1 == n2:
            if operation == CEQUAL_STR:
                length += 1
            else:
                if length:
                    assert operation == CDIFF_STR
                    result.append("%d%s" % (length, CDIFF_STR))
                length, operation = 1, CEQUAL_STR
        else:
            if operation == CDIFF_STR:
                length += 1
            else:
                if length:
                    assert operation == CEQUAL_STR
                    result.append("%d%s" % (length, CEQUAL_STR))
                length, operation = 1, CDIFF_STR

    # Append the final operation.
    result.append("%d%s" % (length, operation))

    return "".join(result)


def makeCigar(reference, query, noEdgeInsertions=True):
    """
    Make a CIGAR string from an aligned reference and query.

    @param reference: A C{str} reference sequence, possibly padded on
        the left and right with spaces, and possibly with '-' characters
        to indicate that the query has an insertion (relative to the
        reference).
    @param query: A C{str} reference sequence, possibly padded on
        the left with spaces, and possibly with '-' characters to indicate
        that the query has a deletion (relative to the reference).
    @param noEdgeInsertions: If {True}, convert a leading and trailing
        insertions in the result into an equivalent number of soft clips.
        This is useful when making a CIGAR string that would otherwise have
        insertions due to the reference containing '-' characters.
    @raise ValueError: If the query or reference is empty.
    @return: A C{str} CIGAR string.
    """
    if not reference:
        raise ValueError("Empty reference")
    if not query:
        raise ValueError("Empty query")

    # Pad the reference on the right, if necessary.
    if len(reference) < len(query):
        reference += " " * (len(query) - len(reference))

    cigar = []
    softClipLeft = softClipRight = 0
    start = True

    for referenceBase, queryBase in zip(reference, query):
        if referenceBase == " ":
            if queryBase == " ":
                continue
            else:
                if start:
                    softClipLeft += 1
                else:
                    softClipRight += 1
        else:
            start = False
            if queryBase == " ":
                continue
            elif referenceBase == "-":
                # Insertion to the reference.
                assert queryBase != "-"
                cigar.append(CINS_STR)
            elif queryBase == "-":
                # Deletion from the reference.
                assert referenceBase != "-"
                cigar.append(CDEL_STR)
            else:
                cigar.append(CMATCH_STR)

    # Replace strings of identical operations with a count and the operation.
    lastOp = None
    count = 0
    middle = []
    for op in cigar:
        if op != lastOp:
            if count:
                middle.append(f"{count}{lastOp}")
            count = 1
            lastOp = op
        else:
            count += 1

    if count:
        middle.append(f"{count}{lastOp}")

    if noEdgeInsertions:
        if middle:
            if middle[0].endswith(CINS_STR):
                softClipLeft += int(middle[0][:-1])
                middle.pop(0)

        if middle:
            if middle[-1].endswith(CINS_STR):
                softClipRight += int(middle[-1][:-1])
                middle.pop()

    return (
        (f"{softClipLeft}S" if softClipLeft else "")
        + "".join(middle)
        + (f"{softClipRight}S" if softClipRight else "")
    )


def cigarTuplesToOperations(tuples, includeHardClip=True):
    """
    Produce a sequence of CIGAR operations given a cigar tuples list
    from a pysam AlignedRead instance.

    See https://pysam.readthedocs.io/en/latest/
        api.html#pysam.AlignedSegment.cigartuples

    @param tuples: A C{list} of 2-tuples, each containing an operation
        and an C{int} count.
    @param includeHardClip: If C{True}, yield hard clipping operations,
        else elide them.
    """
    for operation, count in tuples:
        if operation != CHARD_CLIP or includeHardClip:
            for _ in range(count):
                yield operation


def softClippedOffset(offset, pairs, cigarOperations):
    """
    Get the offset in the reference of a base that has been soft-clipped
    in a query.

    @param offset: The offset (in C{pairs}) of the soft-clipped base (which
       has a to-be-determined offset in the reference).
    @param pairs: A C{list} of C{int} (queryOffset, referenceOffset) pairs
       for a pysam C{pysam.AlignedSegment}, as returned by
       C{read.get_aligned_pairs} where the read is returned by the pysam
       C{fetch} method.
    @param cigarOperations: A C{list} of pysam CIGAR operations.
    @return: The C{int} reference offset of the soft-clipped base.
    """
    assert cigarOperations[offset] == CSOFT_CLIP
    assert pairs[offset][1] is None

    # Look back.
    count = 0
    for pair, cigarOperation in zip(pairs[offset::-1], cigarOperations[offset::-1]):
        if cigarOperation == CSOFT_CLIP:
            count += 1
        else:
            _, referenceOffset = pair
            # If we have a reference offset, then we've found our
            # answer. Otherwise, this has to be a hard-clipped base.
            if referenceOffset is None:
                assert cigarOperation == CHARD_CLIP
                # We could break but won't, as explained above.
            else:
                return referenceOffset + count

    # Look ahead.
    count = 0
    for pair, cigarOperation in zip(pairs[offset:], cigarOperations[offset:]):
        if cigarOperation == CSOFT_CLIP:
            count += 1
        else:
            _, referenceOffset = pair
            # If we have a reference offset, then we've found our
            # answer. Otherwise, this has to be a hard-clipped base.
            if referenceOffset is None:
                assert cigarOperation == CHARD_CLIP
                # We could break here, but if we keep going we can test
                # that the whole rest of the CIGAR is hard clips and will
                # fail if not. That's not needed in a world of perfect code.
            else:
                return referenceOffset - count

    # This should be impossible.
    raise ValueError(
        "Soft-clipped base with no following or preceding non-hard-clipped bases."
    )


def insertionOffset(offset, pairs, cigarOperations):
    """
    Get the insertion offset in the reference for a base that is part of an
    insertion.

    @param offset: The offset (in C{pairs}) of the soft-clipped base (which
       has a to-be-determined offset in the reference).
    @param pairs: A C{list} of C{int} (queryOffset, referenceOffset) pairs
       for a pysam C{pysam.AlignedSegment}, as returned by
       C{read.get_aligned_pairs} where the read is returned by the pysam
       C{fetch} method.
    @param cigarOperations: A C{list} of pysam CIGAR operations.
    @return: A 2-tuple containing a C{bool} and the C{int} reference offset of
        the base after the insertion. The C{bool} indicates whether we needed
        to look back in the sequence to find a reference base.
    """
    assert cigarOperations[offset] == CINS
    assert pairs[offset][1] is None

    # Look back.
    for pair, cigarOperation in zip(pairs[offset::-1], cigarOperations[offset::-1]):
        if cigarOperation != CINS:
            _, referenceOffset = pair
            # If we have a reference offset, then we've found our
            # answer. Otherwise, check as above.
            if referenceOffset is None:
                assert cigarOperation not in CONSUMES_REFERENCE
            else:
                return True, referenceOffset + 1

    # Look ahead.
    for pair, cigarOperation in zip(pairs[offset:], cigarOperations[offset:]):
        if cigarOperation != CINS:
            _, referenceOffset = pair
            # If we have a reference offset, then we've found our
            # answer. Otherwise, check that this is something that does not
            # consume reference bases.
            if referenceOffset is None:
                assert cigarOperation not in CONSUMES_REFERENCE
            else:
                return False, referenceOffset

    # This should be impossible.
    raise ValueError("Inserted base with no following or preceding reference bases.")



import builtins
import os
from io import BytesIO, StringIO
from unittest import TestCase
from unittest.mock import mock_open, patch

from Bio import bgzf

from dark.fasta import (
    FastaReads,
    SqliteIndex,
    combineReads,
    dedupFasta,
    dePrefixAndSuffixFasta,
    fastaSubtract,
)
from dark.reads import AARead, DNARead, Read, Reads, RNARead


class FastaDeDup(TestCase):
    """
    Tests for de-duping FASTA sequence lists.
    """

    def testEmpty(self):
        """
        An empty FASTA list gets de-duped to an empty list.
        """
        self.assertEqual(list(dedupFasta([])), [])

    def testLengthOne(self):
        """
        A FASTA list with just one item gets de-duped to the same one item.
        """
        reads = Reads()
        reads.add(Read("id", "GGG"))
        self.assertEqual(list(dedupFasta(reads)), [Read("id", "GGG")])

    def testRemovalOfIdenticalSequences(self):
        """
        A list with 2 copies of the same seq is de-duped to have 1 copy.
        """
        reads = Reads()
        reads.add(Read("id", "GGG"))
        reads.add(Read("id", "GGG"))
        self.assertEqual(list(dedupFasta(reads)), [Read("id", "GGG")])

    def testRemovalOfIdenticalSequencesWithDifferingIds(self):
        """
        A list with 2 copies of the same seq is de-duped to have 1 copy,
        including when the read ids differ.
        """
        reads = Reads()
        reads.add(Read("id1", "GGG"))
        reads.add(Read("id2", "GGG"))
        self.assertEqual(list(dedupFasta(reads)), [Read("id1", "GGG")])


class TestDePrefixAndSuffixFasta(TestCase):
    def testEmpty(self):
        """
        An empty FASTA list gets de-duped to an empty list.
        """
        self.assertEqual(list(dePrefixAndSuffixFasta([])), [])

    def testLengthOne(self):
        """
        A FASTA list with just one item gets de-duped to the same one item.
        """
        read = Read("hey", "agtcagtcagtc")
        self.assertEqual(list(dePrefixAndSuffixFasta([read])), [read])

    def testRemovalOfIdenticalSequences(self):
        """
        A list with 2 copies of the same seq is de-duped to have 1 copy.
        """
        read1 = Read("hey", "agtcagtcagtc")
        read2 = Read("hey", "agtcagtcagtc")
        self.assertEqual(list(dePrefixAndSuffixFasta([read1, read2])), [read1])

    def testRemovalOfPrefix(self):
        """
        A sequence that is a prefix of another is removed.
        """
        read1 = Read("s1", "agtcagtcagtc")
        read2 = Read("s2", "agtcag")
        self.assertEqual(list(dePrefixAndSuffixFasta([read1, read2])), [read1])

    def testRemovalOfSuffix(self):
        """
        A sequence that is a suffix of another is removed.
        """
        read1 = Read("s1", "agtcagtcagtc")
        read2 = Read("s2", "cagtc")
        self.assertEqual(list(dePrefixAndSuffixFasta([read1, read2])), [read1])

    def testRemovalOfPrefixSuffixAndDuplicate(self):
        """
        Prefixes, suffixes, and duplicates should collectively all be removed.
        """
        read1 = Read("s1", "agtcagtcagtc")
        read2 = Read("s2", "agtcagtcagtc")
        read3 = Read("s3", "agtcagt")
        read4 = Read("s4", "tcagtc")
        self.assertEqual(
            list(dePrefixAndSuffixFasta([read1, read2, read3, read4])), [read1]
        )

    def testOrderIndependent(self):
        """
        A sequence that is a prefix of another is removed when it appears
        first.
        """
        read1 = Read("s1", "agtcag")
        read2 = Read("s2", "agtcagtcagtc")
        self.assertEqual(list(dePrefixAndSuffixFasta([read1, read2])), [read2])


class TestFastaSubtract(TestCase):
    """
    Test the fastaSubtract() function.
    """

    def testZeroFiles(self):
        self.assertRaises(IndexError, fastaSubtract, [])

    def testOneFile(self):
        """
        When just one file is passed we should get a result that has as many
        reads as was in the single input file.
        """
        fasta1 = b"\n".join(
            [
                b">one",
                b"agtcagtcagtc",
                b">two",
                b"acctg",
                b">three",
                b"atgggtc",
                b">four",
                b"atggctattgaactgtatct",
            ]
        )

        result = list(fastaSubtract([BytesIO(fasta1)]))
        self.assertEqual(len(result), 4)

    def testSubtractEverything(self):
        """
        When two input files have the same reads, subtraction must result in an
        empty (no reads) output.
        """
        fasta1 = b"\n".join(
            [
                b">one",
                b"agtcagtcagtc",
                b">two",
                b"acctg",
                b">three",
                b"atgggtc",
                b">four",
                b"atggctattgaactgtatct",
            ]
        )

        result = list(fastaSubtract([BytesIO(fasta1), BytesIO(fasta1)]))
        self.assertEqual([], result)

    def testSubtractFromNothing(self):
        """
        When the first file is empty, the result shoud be too.
        """
        fasta1 = b""
        fasta2 = b"\n".join(
            [
                b">five",
                b"agtcagtcagtc",
                b">six",
                b"acctg",
            ]
        )

        result = list(fastaSubtract([BytesIO(fasta1), BytesIO(fasta2)]))
        self.assertEqual([], result)

    def testSubtractNothing(self):
        """
        When two input files have no overlap, subtraction must result in the
        same reads as are in the first input.
        """
        fasta1 = b"\n".join(
            [
                b">one",
                b"agtcagtcagtc",
                b">two",
                b"acctg",
                b">three",
                b"atgggtc",
                b">four",
                b"atggctattgaactgtatct",
            ]
        )
        fasta2 = b"\n".join(
            [
                b">five",
                b"agtcagtcagtc",
                b">six",
                b"acctg",
            ]
        )

        result = list(fastaSubtract([BytesIO(fasta1), BytesIO(fasta2)]))
        self.assertEqual(
            ["four", "one", "three", "two"], sorted([seq.id for seq in result])
        )

    def testThreeFiles(self):
        """
        Subtraction of three files must work correctly.
        """
        fasta1 = b"\n".join(
            [
                b">one",
                b"agtcagtcagtc",
                b">two",
                b"acctg",
                b">three",
                b"atgggtc",
                b">four",
                b"atggctattgaactgtatct",
            ]
        )
        fasta2 = b"\n".join(
            [
                b">one",
                b"agtcagtcagtc",
            ]
        )
        fasta3 = b"\n".join(
            [
                b">two",
                b"acctg",
                b">three",
                b"atgggtc",
            ]
        )

        result = list(
            fastaSubtract([BytesIO(fasta1), BytesIO(fasta2), BytesIO(fasta3)])
        )
        self.assertEqual(len(result), 1)
        self.assertEqual(str(result[0].sequence), "atggctattgaactgtatct")
        self.assertEqual(str(result[0].id), "four")

    def testSequencesAreChecked(self):
        """
        If a two reads with the same id do not have the same sequence,
        an assertion error must be raised.
        """
        fasta1 = b"\n".join(
            [
                b">one",
                b"ag",
            ]
        )
        fasta2 = b"\n".join(
            [
                b">one",
                b"at",
            ]
        )

        self.assertRaises(
            AssertionError, fastaSubtract, [BytesIO(fasta1), BytesIO(fasta2)]
        )


class TestFastaReads(TestCase):
    """
    Tests for the L{dark.fasta.FastaReads} class.
    """

    def testEmpty(self):
        """
        An empty FASTA file results in an empty iterator.
        """
        with patch.object(builtins, "open", mock_open(read_data=b"")):
            reads = FastaReads("filename.fasta")
            self.assertEqual([], list(reads))

    def testOneRead(self):
        """
        A FASTA file with one read must be read properly.
        """
        data = b"\n".join([b">id1", b"ACGT"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastaReads("filename.fasta"))
            self.assertEqual([Read("id1", "ACGT")], reads)

    def testNoQuality(self):
        """
        A FASTA file read must not have any quality information.
        """
        data = b"\n".join([b">id1", b"ACGT"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastaReads("filename.fasta"))
            self.assertEqual(None, reads[0].quality)

    def testTwoReads(self):
        """
        A FASTA file with two reads must be read properly and its
        sequences must be returned in the correct order.
        """
        data = b"\n".join([b">id1", b"ACGT", b">id2", b"TGCA"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastaReads("filename.fasta"))
            self.assertEqual(2, len(reads))
            self.assertEqual([Read("id1", "ACGT"), Read("id2", "TGCA")], reads)

    def testTypeDefaultsToDNA(self):
        """
        A FASTA file whose type is not specified must result in reads that
        are instances of DNARead.
        """
        data = b"\n".join([b">id1", b"ACGT"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastaReads("filename.fasta"))
            self.assertTrue(isinstance(reads[0], DNARead))

    def testTypeAA(self):
        """
        A FASTA file whose read class is AARead must result in reads that
        are instances of AARead.
        """
        data = b"\n".join([b">id1", b"ACGST"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastaReads("filename.fasta", AARead))
            self.assertTrue(isinstance(reads[0], AARead))

    def testTypeDNA(self):
        """
        A FASTA file whose read class is DNARead must result in reads that
        are instances of DNARead.
        """
        data = b"\n".join([b">id1", b"ACGT"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastaReads("filename.fasta", DNARead))
            self.assertTrue(isinstance(reads[0], DNARead))

    def testTypeRNA(self):
        """
        A FASTA file whose read class is RNARead must result in reads that
        are instances of RNARead.
        """
        data = b"\n".join([b">id1", b"ACGT"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastaReads("filename.fasta", RNARead))
            self.assertTrue(isinstance(reads[0], RNARead))

    def testConvertLowerToUpperCaseIfSpecifiedAARead(self):
        """
        A read needs to be converted from lower to upper case if specified.
        """
        data = b"\n".join([b">id1", b"actgs"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastaReads("filename.fasta", readClass=AARead, upperCase=True))
            self.assertEqual([AARead("id1", "ACTGS")], reads)

    def testConvertLowerToUpperCaseIfSpecifiedDNARead(self):
        """
        A read needs to be converted from lower to upper case if specified.
        """
        data = b"\n".join([b">id1", b"actg"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastaReads("filename.fasta", upperCase=True))
            self.assertEqual([AARead("id1", "ACTG")], reads)

    def testDontConvertLowerToUpperCaseIfNotSpecified(self):
        """
        A read must not be converted from lower to upper case if not specified.
        """
        data = b"\n".join([b">id1", b"actgs"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(FastaReads("filename.fasta", readClass=AARead))
            self.assertEqual([AARead("id1", "actgs")], reads)

    def testFilterRandomSubsetOfZeroFromZeroReads(self):
        """
        It must be possible to select a random subset of zero reads from a set
        of zero reads, where the read count is provided to C{filter} via the
        C{trueLength} argument.
        """
        data = b""
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = FastaReads("filename.fasta")
            result = list(reads.filter(randomSubset=0, trueLength=0))
            self.assertEqual([], result)

    def testFilterRandomSubsetOfTwoFromTwoReads(self):
        """
        It must be possible to select a random subset of two reads from a set
        of two reads, where the read count is provided to C{filter} via the
        C{trueLength} argument.
        """
        data = b"\n".join([b">id1", b"ACGT", b">id2", b"TGCA"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = FastaReads("filename.fasta")
            result = list(reads.filter(randomSubset=2, trueLength=2))
            self.assertEqual([Read("id1", "ACGT"), Read("id2", "TGCA")], result)

    def testFilterRandomSubsetOfOneFromTenReads(self):
        """
        It must be possible to select a random subset of one read from a set
        of ten reads, where the read count is provided to C{filter} via the
        C{trueLength} argument.
        """
        data = b"\n".join([b">id", b"ACGT"] * 10)
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = FastaReads("filename.fasta")
            result = list(reads.filter(randomSubset=1, trueLength=10))
            self.assertEqual(1, len(result))

    def testTwoFiles(self):
        """
        It must be possible to read from two FASTA files.
        """

        class SideEffect:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("file1.fasta", filename)
                    self.count += 1
                    return BytesIO(b">id1\nACTG\n")
                elif self.count == 1:
                    self.test.assertEqual("file2.fasta", filename)
                    self.count += 1
                    return BytesIO(b">id2\nCAGT\n")
                else:
                    self.test.fail("We are only supposed to be called twice!")

        sideEffect = SideEffect(self)
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect.sideEffect
            reads = FastaReads(["file1.fasta", "file2.fasta"])
            self.assertEqual(
                [
                    DNARead("id1", "ACTG"),
                    DNARead("id2", "CAGT"),
                ],
                list(reads),
            )


class TestCombineReads(TestCase):
    """
    Tests for the L{dark.fasta.combineReads} function.
    """

    def testNoneNone(self):
        """
        A C{None} FASTA file name and None sequences results in an empty
        FastaReads instance.
        """
        reads = combineReads(None, None)
        self.assertEqual([], list(reads))

    def testNoneEmpty(self):
        """
        A C{None} FASTA file name and an empty sequences list results in an
        empty FastaReads instance.
        """
        reads = list(combineReads(None, []))
        self.assertEqual([], reads)

    def testFileOnly(self):
        """
        If a FASTA file is given but sequences is C{None}, the resulting
        FastaReads must contain the expected read.
        """
        data = b"\n".join([b">id1", b"ACGT"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(combineReads("filename.fasta", None))
            self.assertEqual([Read("id1", "ACGT")], reads)

    def testNoUpperCaseFileOnly(self):
        """
        If upperCase is not passed and a FASTA file is given, the resulting
        FastaReads must contain the expected read, in the original case.
        """
        data = b"\n".join([b">id1", b"AcgT"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(combineReads("filename.fasta", None))
            self.assertEqual([Read("id1", "AcgT")], reads)

    def testUpperCaseFileOnly(self):
        """
        When passing upperCase=True and a FASTA file, the resulting
        FastaReads must have the read sequence in uppper case.
        """
        data = b"\n".join([b">id1", b"acgt"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(combineReads("filename.fasta", None, upperCase=True))
            self.assertEqual([Read("id1", "ACGT")], reads)

    def testSequencesOnly(self):
        """
        A C{None} FASTA file name and a non-empty sequences list results in a
        FastaReads instance with the expected read.
        """
        reads = list(combineReads(None, ["id ACGTSSS"], readClass=AARead))
        self.assertEqual([AARead("id", "ACGTSSS")], reads)

    def testNoUpperCaseSequencesOnly(self):
        """
        If upperCase is not passed to combineReads the resulting read
        sequences must have their original case.
        """
        reads = list(combineReads(None, ["id aCGt"]))
        self.assertEqual([Read("id", "aCGt")], reads)

    def testUpperCaseSequencesOnly(self):
        """
        Passing upperCase=True to combineReads must result in read sequences
        being upper cased.
        """
        reads = list(combineReads(None, ["id acgt"], upperCase=True))
        self.assertEqual([Read("id", "ACGT")], reads)

    def testDefaultReadIdPrefix(self):
        """
        A C{None} FASTA file name and a non-empty sequences list with a
        sequence that has no id results in a FastaReads instance with the
        expected read.
        """
        reads = list(combineReads(None, ["ACGT"]))
        self.assertEqual([Read("command-line-read-1", "ACGT")], reads)

    def testCustomReadIdPrefix(self):
        """
        A C{None} FASTA file name and a non-empty sequences list with a
        sequence that has no id, but with a custom read id prefix, results in a
        FastaReads instance with the expected read.
        """
        reads = list(
            combineReads(None, ["ACGTSSS"], idPrefix="prefix-", readClass=AARead)
        )
        self.assertEqual([AARead("prefix-1", "ACGTSSS")], reads)

    def testSpecificReadClass(self):
        """
        A specific read class must result in a FastaReads instance with reads
        of that class, both for reads from a FASTA file and from individually
        specified sequences.
        """
        data = b"\n".join([b">id1", b"ACGT"])
        with patch.object(builtins, "open", mock_open(read_data=data)):
            reads = list(combineReads("filename.fasta", ["ACGT"], readClass=RNARead))
            self.assertTrue(isinstance(reads[0], RNARead))
            self.assertTrue(isinstance(reads[1], RNARead))


class TestSqliteIndex(TestCase):
    """
    Tests for the SqliteIndex class.
    """

    def testAddFilename(self):
        """
        Test the internal _addFilename method.
        """
        index = SqliteIndex(":memory:")
        self.assertEqual(1, index._addFilename("filename1.fasta"))
        self.assertEqual(2, index._addFilename("filename2.fasta"))
        index.close()

    def testAddDuplicateFilename(self):
        """
        When _addFilename is called twice with the same name, a ValueError
        must be raised.
        """
        index = SqliteIndex(":memory:")
        self.assertEqual(1, index._addFilename("f.fas"))
        error = "^Duplicate file name: 'f.fas'$"
        self.assertRaisesRegex(ValueError, error, index._addFilename, "f.fas")

    def testGetNonexistentFilename(self):
        """
        If the internal _getFilename method is called with a file number that
        has not been added, it must return None.
        """
        index = SqliteIndex(":memory:")
        self.assertEqual(None, index._getFilename(1))
        index.close()

    def testGetFilename(self):
        """
        The internal _getFilename method must return the expected result.
        """
        index = SqliteIndex(":memory:")
        self.assertEqual(1, index._addFilename("filename.fasta"))
        self.assertEqual("filename.fasta", index._getFilename(1))
        index.close()

    def testGetNonexistentFileNumber(self):
        """
        If the internal _getFileNumber method is called with a file whose name
        has not been added, it must return None.
        """
        index = SqliteIndex(":memory:")
        self.assertEqual(None, index._getFileNumber("filename.fasta"))
        index.close()

    def testGetFileNumber(self):
        """
        The internal _getFileNumber method must return the expected result.
        """
        index = SqliteIndex(":memory:")
        self.assertEqual(1, index._addFilename("filename.fasta"))
        self.assertEqual(1, index._getFileNumber("filename.fasta"))
        index.close()

    def testBZ2File(self):
        """
        Trying to add a .bz2 file must result in a ValueError.
        """
        index = SqliteIndex(":memory:")
        error = (
            "^Compressed FASTA is only supported in BGZF format\\. Use "
            "bgzip to compresss your FASTA\\.$"
        )
        self.assertRaisesRegex(ValueError, error, index.addFile, "file.bz2")

    def testAddOneFile(self):
        """
        Test the creation of an index with sequences added from one file.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("filename.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\n>id2\nAACCTTGG\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            self.assertEqual(2, index.addFile("filename.fasta"))
            index.close()

    def testAddFileWithDuplicateSequence(self):
        """
        If a sequence id is duplicated in a FASTA file, a ValueError must be
        raised.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("filename.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\n>id1\nAACCTTGG\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            error = "^FASTA sequence id 'id1' found twice in file 'filename.fasta'\\.$"
            self.assertRaisesRegex(ValueError, error, index.addFile, "filename.fasta")
            index.close()

    def testAddFilesWithDuplicateSequence(self):
        """
        If a sequence id occurs in more than one FASTA file, a ValueError must
        be raised.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("filename1.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\n>id2\nAACCTTGG\n")
                elif self.count == 1:
                    self.test.assertEqual("filename2.fasta", filename)
                    self.count += 1
                    return StringIO(">id2\nAAACCC\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            index.addFile("filename1.fasta")
            error = (
                "^FASTA sequence id 'id2', found in file "
                "'filename2\\.fasta', was previously added from file "
                "'filename1\\.fasta'\\.$"
            )
            self.assertRaisesRegex(ValueError, error, index.addFile, "filename2.fasta")
            index.close()

    def testAddDuplicateFile(self):
        """
        If a filename is passed to addFile more than once, a ValueError must
        be raised.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("filename.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\n>id2\nAACCTTGG\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            self.assertEqual(2, index.addFile("filename.fasta"))
            error = "^Duplicate file name: 'filename\\.fasta'$"
            self.assertRaisesRegex(
                ValueError, error, index._addFilename, "filename.fasta"
            )
            index.close()

    def testFind(self):
        """
        The _find method must return the expected filename and offset.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("filename.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\n>id2\nAACCTTGG\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            index.addFile("filename.fasta")
            self.assertEqual(("filename.fasta", 5), index._find("id1"))
            self.assertEqual(("filename.fasta", 15), index._find("id2"))
            index.close()

    def testFindWithTwoFiles(self):
        """
        The _find method must return the expected filename and offset when
        sequences are added from two files.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("filename1.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\n>id2\nAACCTTGG\n")
                elif self.count == 1:
                    self.test.assertEqual("filename2.fasta", filename)
                    self.count += 1
                    return StringIO(">sequence3\nAAACCC\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            index.addFile("filename1.fasta")
            index.addFile("filename2.fasta")
            self.assertEqual(("filename1.fasta", 5), index._find("id1"))
            self.assertEqual(("filename1.fasta", 15), index._find("id2"))
            self.assertEqual(("filename2.fasta", 11), index._find("sequence3"))
            index.close()

    def testDictLookupSequenceCrossesNewlines(self):
        """
        The __getitem__ method (i.e., dictionary-like lookup) must return the
        expected read when the sequence spans multiple lines of the input file,
        including lines ending in \n and \r\n.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0 or self.count == 1:
                    self.test.assertEqual("filename.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\r\nCCCC\nGGG\n>id2\nAACCTG\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            index.addFile("filename.fasta")
            self.assertEqual(DNARead("id1", "ACTGCCCCGGG"), index["id1"])
            index.close()

    def testDictLookupWithFastaDirectory(self):
        """
        The __getitem__ method (i.e., dictionary-like lookup) must return the
        expected read, obtained from the expected file name, when a FASTA base
        directory is specified.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0:
                    self.test.assertEqual("/tmp/f.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\r\nCCCC\nGGG\n>id2\nAACCTG\n")
                if self.count == 1:
                    self.test.assertEqual(
                        os.path.join("/usr/local/fasta", "f.fasta"), filename
                    )
                    self.count += 1
                    return StringIO(">id1\nACTG\r\nCCCC\nGGG\n>id2\nAACCTG\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:", fastaDirectory="/usr/local/fasta")
            index.addFile("/tmp/f.fasta")
            self.assertEqual(DNARead("id1", "ACTGCCCCGGG"), index["id1"])
            index.close()

    def testDictLookupSequenceLastInFile(self):
        """
        The __getitem__ method (i.e., dictionary-like lookup) must return the
        expected read when the sequence spans multiple lines and is the last
        one in the input file.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0 or self.count == 1:
                    self.test.assertEqual("filename.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\r\nCCCC\n>id2\nAACCTG\nAAA\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            index.addFile("filename.fasta")
            self.assertEqual(DNARead("id2", "AACCTGAAA"), index["id2"])
            index.close()

    def testDictLookupSequenceMiddleOfThree(self):
        """
        The __getitem__ method (i.e., dictionary-like lookup) must return the
        expected read when the sequence spans multiple lines and is the middle
        one of three sequences in the input file.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0 or self.count == 1:
                    self.test.assertEqual("filename.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\nCCCC\n>id2\nAACCTG\nAAA\n>id3\nAAA\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            index.addFile("filename.fasta")
            self.assertEqual(DNARead("id2", "AACCTGAAA"), index["id2"])
            index.close()

    def testDictLookupWithTwoFiles(self):
        """
        The __getitem__ method (i.e., dictionary-like lookup) must return the
        expected reads when sequences are added from two files.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0 or self.count == 2 or self.count == 3:
                    self.test.assertEqual("filename1.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nACTG\n>id2\nAACCTTGG\n")
                elif self.count == 1 or self.count == 4:
                    self.test.assertEqual("filename2.fasta", filename)
                    self.count += 1
                    return StringIO(">seq3\nAAACCC\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            index.addFile("filename1.fasta")
            index.addFile("filename2.fasta")
            self.assertEqual(DNARead("id1", "ACTG"), index["id1"])
            self.assertEqual(DNARead("id2", "AACCTTGG"), index["id2"])
            self.assertEqual(DNARead("seq3", "AAACCC"), index["seq3"])
            index.close()

    def testDictLookupSpecificReadClass(self):
        """
        The __getitem__ method (i.e., dictionary-like lookup) must return the
        expected read type.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count == 0 or self.count == 1:
                    self.test.assertEqual("filename.fasta", filename)
                    self.count += 1
                    return StringIO(">id1\nMM\n>id2\n")
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(builtins, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:", readClass=AARead)
            index.addFile("filename.fasta")
            result = index["id1"]
            self.assertTrue(isinstance(result, AARead))
            self.assertEqual(AARead("id1", "MM"), result)
            index.close()

    def testDictLookupGzipDataWithBGZsuffix(self):
        """
        The __getitem__ method (i.e., dictionary-like lookup) must return the
        expected read when the index file is in BGZF format and has a .bgz
        suffix.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count <= 1:
                    self.test.assertEqual("filename.fasta.bgz", filename)
                    self.count += 1
                    writerIO = BytesIO()
                    writer = bgzf.BgzfWriter(fileobj=writerIO)
                    writer.write(b">id0\nAC\n")
                    writer.flush()
                    fileobj = BytesIO(writerIO.getvalue())
                    return bgzf.BgzfReader(fileobj=fileobj)
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(bgzf, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            index.addFile("filename.fasta.bgz")
            self.assertEqual(DNARead("id0", "AC"), index["id0"])
            index.close()

    def testDictLookupGzipData(self):
        """
        The __getitem__ method (i.e., dictionary-like lookup) must return the
        expected reads when sequences span multiple lines of the input file,
        and include lines ending in \n and \r\n and have been compressed with
        bgzip, including when sequences are more than 64K bytes into the input
        file.
        """

        class Open:
            def __init__(self, test):
                self.test = test
                self.count = 0

            def sideEffect(self, filename, *args, **kwargs):
                if self.count <= 4:
                    self.test.assertEqual("filename.fasta.gz", filename)
                    self.count += 1
                    writerIO = BytesIO()
                    writer = bgzf.BgzfWriter(fileobj=writerIO)
                    writer.write(
                        b">id0\nAC\n"
                        + b">id1\n"
                        + (b"A" * 70000)
                        + b"\n"
                        + b">id2\r\nACTG\r\nCCCC\r\nGGG\r\n"
                        + b">id3\nAACCTG\n"
                    )
                    writer.flush()
                    fileobj = BytesIO(writerIO.getvalue())
                    return bgzf.BgzfReader(fileobj=fileobj)
                else:
                    self.test.fail(
                        "Open called too many times. Filename: %r, Args: %r, "
                        "Keyword args: %r." % (filename, args, kwargs)
                    )

        sideEffect = Open(self).sideEffect
        with patch.object(bgzf, "open") as mockMethod:
            mockMethod.side_effect = sideEffect
            index = SqliteIndex(":memory:")
            index.addFile("filename.fasta.gz")
            self.assertEqual(DNARead("id0", "AC"), index["id0"])
            self.assertEqual(DNARead("id1", "A" * 70000), index["id1"])
            self.assertEqual(DNARead("id2", "ACTGCCCCGGG"), index["id2"])
            self.assertEqual(DNARead("id3", "AACCTG"), index["id3"])
            index.close()



#!/usr/bin/env python

import argparse
import sys
from typing import Optional

from ete4 import Tree

from dark.trees import ete4root


def parseArgs():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Curate a phylogenetic tree using the ete4 library and print the result."
        ),
    )

    parser.add_argument("treeFile", help="The tree file to curate.")

    parser.add_argument(
        "--parser",
        "--format",
        type=int,
        default=0,
        help=(
            "The Newick tree flavor. See "
            "https://etetoolkit.github.io/ete/reference/reference_parsers.html "
            "for information on the Newick variations ete4 understands. The output "
            "tree will be in the same format as the input."
        ),
    )

    parser.add_argument(
        "--rootNode",
        action="append",
        help=(
            "Re-root the tree on the branch entering the node that is the "
            "most-recent common ancestor of the tips with these names (may be "
            "repeated)."
        ),
    )

    parser.add_argument(
        "--scale",
        default=1.0,
        type=float,
        help=(
            "Factor to scale (i.e., divide) the length of the branches leading to the "
            "descendants of the new root (if --rootNode is used). This can be used to "
            "reduce the length of very long branches that result from rerooting on a "
            "distant outgroup. If you use this option for an image in a publication, "
            "you of course must point out that you have done so!"
        ),
    )

    parser.add_argument(
        "--detach",
        action="store_true",
        help=(
            "Detach the named root node from the tree after rooting on the edge coming "
            "into it. The newly added (by rooting) parent of the named node will be "
            "the (temporary) new root node. This option will first remove the named "
            "node as a child of that root, leaving it with just one child. The "
            "remaining child will then be made the new root.  This essentially like "
            "rooting on an outgroup taxon then completely removing the outgroup from "
            "the tree."
        ),
    )

    parser.add_argument(
        "--collapse",
        action="store_true",
        help="Collapse branches that are insufficiently supported.",
    )

    parser.add_argument(
        "--supportThreshold",
        type=int,
        help=(
            "The support level below which edges will be collapsed (only valid if "
            "--collapse is used)."
        ),
    )

    parser.add_argument(
        "--alrtThreshold",
        type=float,
        help=(
            "The alrt level below which edges will be collapsed (only valid if "
            "--collapse is used). Alrt values are added by iqtree2 if you use its "
            "-alrt option and can be used to check for well-supported branches."
        ),
    )

    parser.add_argument(
        "--removeSupport",
        action="store_true",
        help="Remove support values from internal nodes.",
    )

    parser.add_argument(
        "--removeAlrt",
        action="store_true",
        help="Remove alrt values from internal nodes.",
    )

    parser.add_argument(
        "--verbose",
        action="store_true",
        help="Print information about processing to standard error.",
    )

    return parser.parse_args()


def processSupport(tree: Tree, args: argparse.Namespace) -> None:
    """
    Potentially collapse and change the labels on internal nodes.
    """
    support: Optional[int]
    alrt: Optional[float]

    for node in tree.traverse(strategy="postorder"):
        if node.is_leaf:
            continue

        if node.name:
            try:
                alrts, supports = node.name.split("/")
            except ValueError:
                support = int(node.name)
                alrt = None
            else:
                alrt = float(alrts)
                support = int(supports)
        else:
            support = alrt = None

        if args.collapse:
            doCollapse = (
                args.supportThreshold is not None
                and support is not None
                and support < args.supportThreshold
            ) or (
                args.alrtThreshold is not None
                and alrt is not None
                and alrt < args.alrtThreshold
            )

            if doCollapse:
                if args.verbose:
                    print(f"Collapsing node with label {node.name}", file=sys.stderr)
                for child in node.children:
                    child.dist += node.dist

                node.delete()

        if args.removeSupport:
            if args.removeAlrt:
                name = ""
            else:
                name = "" if alrt is None else alrts
        else:
            if args.removeAlrt:
                name = supports
            else:
                name = node.name

        if node.name != name:
            if args.verbose:
                print(
                    f"Changing internal node label {node.name!r} -> {name!r}",
                    file=sys.stderr,
                )
            node.name = name


def main() -> None:
    args = parseArgs()
    tree = Tree(args.treeFile, parser=args.parser)

    processSupport(tree, args)

    if args.rootNode:
        tree = ete4root(tree, args.rootNode, args.detach, args.scale)

    print(tree.write(parser=args.parser))


if __name__ == "__main__":
    main()
